r"""Test basic network blocks."""

import pytest

import torch
from torch import Tensor, nn

from deepali.networks.blocks import DenseBlock
from deepali.networks.blocks import ResidualUnit
from deepali.networks.blocks import SkipConnection
from deepali.networks.layers import ConvLayer, JoinLayer, LambdaLayer


def test_dense_block() -> None:
    r"""Test block with dense skip connections."""

    x = torch.tensor([[1.0, -2.0, 3.0, 4.0, -5.0]])

    scale = LambdaLayer(lambda a: 2 * a)
    square = LambdaLayer(lambda a: a.square())
    clamp = LambdaLayer(lambda a: a.clamp(min=0))

    block = DenseBlock(scale, square, clamp)
    assert isinstance(block.layers, nn.ModuleDict)
    assert "0" in block.layers
    assert "1" in block.layers
    assert "2" in block.layers
    assert block.layers["0"] is scale
    assert block.layers["1"] is square
    assert block.layers["2"] is clamp

    block = DenseBlock({"scale": scale, "square": square, "clamp": clamp}, join="concat")
    assert isinstance(block.layers, nn.ModuleDict)
    assert "scale" in block.layers
    assert "square" in block.layers
    assert "clamp" in block.layers
    assert block.layers["scale"] is scale
    assert block.layers["square"] is square
    assert block.layers["clamp"] is clamp

    y = block(x)
    assert isinstance(y, Tensor)

    a = scale(x)
    b = square(torch.cat([x, a], dim=1))
    c = clamp(torch.cat([x, a, b], dim=1))
    assert isinstance(c, Tensor)
    assert y.shape == c.shape
    assert y.allclose(c)

    block = DenseBlock({"scale": scale, "square": square, "clamp": clamp}, join="add")

    y = block(x)
    assert isinstance(y, Tensor)
    assert y.shape == x.shape

    a = scale(x)
    b = square(x + a)
    c = clamp(x + a + b)
    assert isinstance(c, Tensor)
    assert y.shape == c.shape
    assert y.allclose(c)


def test_residual_unit() -> None:
    r"""Test convoluational residual block."""

    with pytest.raises(TypeError):
        ResidualUnit(2, 1, 1, num_layers=1.0)
    with pytest.raises(ValueError):
        ResidualUnit(2, 1, 1, num_layers=0)
    with pytest.raises(ValueError):
        ResidualUnit(2, in_channels=64, out_channels=64, num_channels=32, num_layers=2)

    x = torch.tensor([[[1.0, -2.0, 3.0, 4.0, -5.0]]])

    block = ResidualUnit(spatial_dims=1, in_channels=1, order="nac")
    assert isinstance(block, nn.Module)
    assert isinstance(block, SkipConnection)
    assert isinstance(block.func, nn.Sequential)
    assert len(block.func) == 2
    assert isinstance(block.func[0], ConvLayer)
    assert isinstance(block.func[1], ConvLayer)
    assert isinstance(block.skip, nn.Identity)

    y = block.join([x, x])
    assert isinstance(y, Tensor)
    assert y.eq(x + x).all()

    block = ResidualUnit(spatial_dims=1, in_channels=1, out_channels=2)
    assert block.spatial_dims == 1
    assert block.in_channels == 1
    assert block.out_channels == 2
    assert isinstance(block.skip, nn.Conv1d)
    assert block.skip.out_channels == 2

    block = ResidualUnit(
        spatial_dims=1,
        in_channels=1,
        out_channels=2,
        acti="relu",
        norm="group",
        num_layers=1,
        order="nac",
    )
    assert len(block.func) == 1
    assert isinstance(block.func[0].conv, nn.Conv1d)
    assert isinstance(block.func[0].norm, nn.GroupNorm)
    assert isinstance(block.func[0].acti, nn.ReLU)
    assert block.func[0].order == "NAC"
    assert not isinstance(block.join, nn.Sequential)

    block = ResidualUnit(
        spatial_dims=1,
        in_channels=1,
        out_channels=2,
        acti="relu",
        norm="group",
        num_layers=3,
        order="cna",
    )
    assert len(block.func) == 3
    assert isinstance(block.func[0].conv, nn.Conv1d)
    assert isinstance(block.func[0].norm, nn.GroupNorm)
    assert isinstance(block.func[0].acti, nn.ReLU)
    assert isinstance(block.func[1].conv, nn.Conv1d)
    assert isinstance(block.func[1].norm, nn.GroupNorm)
    assert isinstance(block.func[1].acti, nn.ReLU)
    assert isinstance(block.func[2].conv, nn.Conv1d)
    assert isinstance(block.func[2].norm, nn.GroupNorm)
    assert block.func[2].acti is None
    assert block.func[0].order == "CNA"
    assert isinstance(block.join, nn.Sequential)
    assert isinstance(block.join[0], JoinLayer)
    assert isinstance(block.join[1], nn.ReLU)

    y = block(x)
    assert isinstance(y, Tensor)
    assert y.shape == (x.shape[0], 2) + x.shape[2:]


def test_skip_connection() -> None:
    r"""Test skip connection block."""

    x = torch.tensor([[1.0, 2.0, 3.0]])
    module = LambdaLayer(lambda a: a)

    # skip: identity, join: cat
    block = SkipConnection(module)
    assert isinstance(block, nn.Module)
    assert isinstance(block.skip, nn.Identity)
    assert block.func is module

    y = block.join([x, x])
    assert isinstance(y, Tensor)
    assert y.shape == (x.shape[0], 2 * x.shape[1])

    y = block(x)
    assert isinstance(y, Tensor)
    assert y.shape == (x.shape[0], 2 * x.shape[1])
    assert y.eq(torch.cat([x, x], dim=1)).all()

    # skip: identity, join: add
    block = SkipConnection(module, name="residual", join="add")
    assert isinstance(block, nn.Module)
    assert isinstance(block.skip, nn.Identity)
    assert block.func is module

    y = block(x)
    assert isinstance(y, Tensor)
    assert y.shape == x.shape
    assert y.eq(x + x).all()

    # skip: identity, join: mul
    block = SkipConnection(module, skip="identity", join="mul")
    assert isinstance(block, nn.Module)
    assert isinstance(block.skip, nn.Identity)
    assert block.func is module

    y = block(x)
    assert isinstance(y, Tensor)
    assert y.shape == x.shape
    assert y.eq(x * x).all()

    # skip: custom, join: cat
    skip = LambdaLayer(lambda b: 2 * b)
    block = SkipConnection(module, skip=skip, join="cat")
    assert block.skip is skip
    assert block.func is module

    y = block(x)
    assert isinstance(y, Tensor)
    assert y.shape == (x.shape[0], 2 * x.shape[1])
    assert y.eq(torch.cat([x, 2 * x], dim=1)).all()



r"""Normalization layers."""

from numbers import Integral
from functools import partial
from typing import Any, Callable, Mapping, Optional, Sequence, Union

from torch import Tensor, nn
from torch.nn import Module

from .lambd import LambdaLayer


NormFunc = Callable[[Tensor], Tensor]
NormArg = Union[NormFunc, str, Mapping[str, Any], Sequence, None]


def normalization(
    arg: NormArg,
    *args,
    spatial_dims: Optional[int] = None,
    num_features: Optional[int] = None,
    **kwargs,
) -> Module:
    r"""Create normalization layer.

    Args:
        arg: Custom normalization function or module, or name of normalization layer with optional keyword arguments.
        args: Positional arguments passed to normalization layer.
        num_features: Number of input features.
        spatial_dims: Number of spatial dimensions of input tensors.
        kwargs: Additional keyword arguments for normalization layer. Overrides keyword arguments given as second
            tuple item when ``arg`` is a ``(name, kwargs)`` tuple instead of a string.

    Returns:
        Given normalization function when ``arg`` is a ``torch.nn.Module``, or a new normalization layer otherwise.

    """
    if isinstance(arg, Module) and not args and not kwargs:
        return arg
    if callable(arg):
        return NormLayer(arg, *args, **kwargs)
    norm_name = "identity"
    norm_args = {}
    if isinstance(arg, str):
        norm_name = arg
    elif isinstance(arg, Mapping):
        norm_name = arg.get("name")
        if not norm_name:
            raise ValueError("normalization() 'arg' map must contain 'name'")
        if not isinstance(norm_name, str):
            raise TypeError("normalization() 'name' must be str")
        norm_args = {key: value for key, value in arg.items() if key != "name"}
    elif isinstance(arg, Sequence):
        if len(arg) != 2:
            raise ValueError("normalization() 'arg' sequence must have length two")
        norm_name, norm_args = arg
        if not isinstance(norm_name, str):
            raise TypeError("normalization() first 'arg' sequence argument must be str")
        if not isinstance(norm_args, dict):
            if norm_name == "group" and isinstance(norm_args, Integral):
                norm_args = dict(num_groups=norm_args)
            else:
                raise TypeError("normalization() second 'arg' sequence argument must be dict")
        norm_args = norm_args.copy()
    elif arg is not None:
        raise TypeError("normalization() 'arg' must be str, mapping, 2-tuple, callable, or None")
    norm_name = norm_name.lower()
    norm_args.update(kwargs)
    if norm_name in ("none", "identity"):
        norm = nn.Identity()
    elif norm_name in ("batch", "batchnorm"):
        if spatial_dims is None:
            raise ValueError("normalization() 'spatial_dims' required for 'batch' norm")
        if spatial_dims < 0 or spatial_dims > 3:
            raise ValueError("normalization() 'spatial_dims' must be 1, 2, or 3")
        if num_features is None:
            raise ValueError("normalization() 'num_features' required for 'batch' norm")
        norm_type = (nn.BatchNorm1d, nn.BatchNorm2d, nn.BatchNorm3d)[spatial_dims - 1]
        norm = norm_type(num_features, *args, **norm_args)
    elif norm_name in ("group", "groupnorm"):
        num_groups = norm_args.pop("num_groups", 1)
        if num_features is None:
            if "num_channels" not in norm_args:
                raise ValueError("normalization() 'num_features' required for 'group' norm")
            num_features = norm_args.pop("num_channels")
        norm = nn.GroupNorm(num_groups, num_features, *args, **norm_args)
    elif norm_name in ("layer", "layernorm"):
        if num_features is None:
            raise ValueError("normalization() 'num_features' required for 'layer' norm")
        # See examples at https://pytorch.org/docs/stable/generated/torch.nn.GroupNorm.html
        # and Figure 2 of "Group Normalization" paper (https://arxiv.org/abs/1803.08494).
        norm = nn.GroupNorm(1, num_features, *args, **norm_args)
    elif norm_name in ("instance", "instancenorm"):
        if spatial_dims is None:
            raise ValueError("normalization() 'spatial_dims' required for 'instance' norm")
        if spatial_dims < 0 or spatial_dims > 3:
            raise ValueError("normalization() 'spatial_dims' must be 1, 2, or 3")
        if num_features is None:
            raise ValueError("normalization() 'num_features' required for 'instance' norm")
        norm_type = (nn.InstanceNorm1d, nn.InstanceNorm2d, nn.InstanceNorm3d)[spatial_dims - 1]
        norm = norm_type(num_features, *args, **norm_args)
    else:
        raise ValueError("normalization() unknown layer type {norm_name!r}")
    return norm


def norm_layer(*args, **kwargs) -> Module:
    r"""Create normalization layer, see ``normalization``."""
    return normalization(*args, **kwargs)


def is_norm_layer(arg: Any) -> bool:
    r"""Whether given module is a normalization layer."""
    if isinstance(arg, NormLayer):
        return True
    return is_batch_norm(arg) or is_group_norm(arg) or is_instance_norm(arg)


def is_batch_norm(arg: Any) -> bool:
    r"""Whether given module is a batch normalization layer."""
    return isinstance(arg, (nn.BatchNorm1d, nn.BatchNorm2d, nn.BatchNorm3d))


def is_group_norm(arg: Any) -> bool:
    r"""Whether given module is a group normalization layer."""
    return isinstance(arg, nn.GroupNorm)


def is_instance_norm(arg: Any) -> bool:
    r"""Whether given module is an instance normalization layer."""
    return isinstance(arg, (nn.InstanceNorm1d, nn.InstanceNorm2d, nn.InstanceNorm3d))


class NormLayer(LambdaLayer):
    r"""Normalization layer."""

    def __init__(
        self,
        arg: NormArg,
        *args,
        spatial_dims: Optional[int] = None,
        num_features: Optional[int] = None,
        **kwargs: Mapping[str, Any],
    ) -> None:
        if callable(arg):
            norm = partial(arg, *args, **kwargs) if args or kwargs else arg
        else:
            kwargs.update(dict(spatial_dims=spatial_dims, num_features=num_features))
            norm = normalization(arg, *args, **kwargs)
        super().__init__(norm)



r"""Flow field regularization terms."""

from __future__ import annotations

from typing import Optional, Union

import torch
from torch import Tensor

from deepali.core.typing import ScalarOrTuple, Shape

from . import functional as L
from .base import DisplacementLoss


class _SpatialDerivativesLoss(DisplacementLoss):
    r"""Base class of regularization terms based on spatial derivatives of dense displacements."""

    def __init__(
        self,
        mode: Optional[str] = None,
        sigma: Optional[float] = None,
        stride: Optional[ScalarOrTuple] = None,
        reduction: str = "mean",
    ):
        r"""Initialize regularization term.

        Args:
            mode: Method used to approximate :func:`flow_derivatives()`.
            sigma: Standard deviation of Gaussian in grid units used to smooth vector field.
            stride: Number of output grid points between control points plus one for ``mode='bspline'``.
            reduction: Operation to use for reducing spatially distributed loss values.

        """
        super().__init__()
        self.mode = mode
        self.sigma = sigma
        self.stride = stride
        self.reduction = reduction

    def _spacing(self, u_shape: Shape) -> Optional[Tensor]:
        ndim = len(u_shape)
        if ndim < 3:
            raise ValueError(f"{type(self).__name__}.forward() 'u' must be at least 3-dimensional")
        if ndim == 3:
            return None
        size = torch.tensor(u_shape[-1:1:-1], dtype=torch.float, device=torch.device("cpu"))
        return 2 / (size - 1)

    def extra_repr(self) -> str:
        args = []
        if self.mode:
            args.append(f"mode={self.mode!r}")
        if self.sigma:
            args.append(f"sigma={self.sigma!r}")
        if self.stride:
            args.append(f"stride={self.stride!r}")
        args.append(f"reduction={self.reduction!r}")
        return ", ".join(args)


class GradLoss(_SpatialDerivativesLoss):
    r"""Displacement field gradient loss."""

    def __init__(
        self,
        p: Union[int, float] = 2,
        q: Optional[Union[int, float]] = 1,
        mode: Optional[str] = None,
        sigma: Optional[float] = None,
        stride: Optional[ScalarOrTuple] = None,
        reduction: str = "mean",
    ):
        r"""Initialize regularization term.

        Args:
            mode: Method used to approximate :func:`flow_derivatives()`.
            sigma: Standard deviation of Gaussian in grid units used to smooth vector field.
            stride: Number of output grid points between control points plus one for ``mode='bspline'``.
            reduction: Operation to use for reducing spatially distributed loss values.

        """
        super().__init__(mode=mode, sigma=sigma, stride=stride, reduction=reduction)
        self.p = p
        self.q = 1 / p if q is None else q

    def forward(self, u: Tensor) -> Tensor:
        r"""Evaluate regularization loss for given transformation."""
        spacing = self._spacing(u.shape)
        return L.grad_loss(
            u,
            p=self.p,
            q=self.q,
            mode=self.mode,
            sigma=self.sigma,
            spacing=spacing,
            stride=self.stride,
            reduction=self.reduction,
        )

    def extra_repr(self) -> str:
        return f"p={self.p}, q={self.q}, " + super().extra_repr()


class Bending(_SpatialDerivativesLoss):
    r"""Bending energy of displacement field."""

    def forward(self, u: Tensor) -> Tensor:
        r"""Evaluate regularization loss for given transformation."""
        spacing = self._spacing(u.shape)
        return L.bending_loss(
            u,
            mode=self.mode,
            sigma=self.sigma,
            spacing=spacing,
            stride=self.stride,
            reduction=self.reduction,
        )


BendingEnergy = Bending
BE = Bending


class Curvature(_SpatialDerivativesLoss):
    r"""Curvature of displacement field."""

    def forward(self, u: Tensor) -> Tensor:
        r"""Evaluate regularization loss for given transformation."""
        spacing = self._spacing(u.shape)
        return L.curvature_loss(
            u,
            mode=self.mode,
            sigma=self.sigma,
            spacing=spacing,
            stride=self.stride,
            reduction=self.reduction,
        )


class Diffusion(_SpatialDerivativesLoss):
    r"""Diffusion of displacement field."""

    def forward(self, u: Tensor) -> Tensor:
        r"""Evaluate regularization loss for given transformation."""
        spacing = self._spacing(u.shape)
        return L.diffusion_loss(
            u,
            mode=self.mode,
            sigma=self.sigma,
            spacing=spacing,
            stride=self.stride,
            reduction=self.reduction,
        )


class Divergence(_SpatialDerivativesLoss):
    r"""Divergence of displacement field."""

    def forward(self, u: Tensor) -> Tensor:
        r"""Evaluate regularization loss for given transformation."""
        spacing = self._spacing(u.shape)
        return L.divergence_loss(
            u,
            mode=self.mode,
            sigma=self.sigma,
            spacing=spacing,
            stride=self.stride,
            reduction=self.reduction,
        )


class Elasticity(_SpatialDerivativesLoss):
    r"""Linear elasticity of displacement field."""

    def __init__(
        self,
        material_name: Optional[str] = None,
        first_parameter: Optional[float] = None,
        second_parameter: Optional[float] = None,
        poissons_ratio: Optional[float] = None,
        youngs_modulus: Optional[float] = None,
        shear_modulus: Optional[float] = None,
        mode: Optional[str] = None,
        sigma: Optional[float] = None,
        stride: Optional[ScalarOrTuple] = None,
        reduction: str = "mean",
    ):
        super().__init__(mode=mode, sigma=sigma, reduction=reduction)
        self.material_name = material_name
        self.first_parameter = first_parameter
        self.second_parameter = second_parameter
        self.poissons_ratio = poissons_ratio
        self.youngs_modulus = youngs_modulus
        self.shear_modulus = shear_modulus

    def forward(self, u: Tensor) -> Tensor:
        r"""Evaluate regularization loss for given transformation."""
        spacing = self._spacing(u.shape)
        return L.elasticity_loss(
            u,
            material_name=self.material_name,
            first_parameter=self.first_parameter,
            second_parameter=self.second_parameter,
            poissons_ratio=self.poissons_ratio,
            youngs_modulus=self.youngs_modulus,
            shear_modulus=self.shear_modulus,
            mode=self.mode,
            sigma=self.sigma,
            spacing=spacing,
            stride=self.stride,
            reduction=self.reduction,
        )

    def extra_repr(self) -> str:
        args = []
        if self.material_name:
            args.append(f"material_name={self.material_name!r}")
        if self.first_parameter is not None:
            args.append(f"first_parameter={self.first_parameter!r}")
        if self.second_parameter is not None:
            args.append(f"second_parameter={self.second_parameter!r}")
        if self.poissons_ratio is not None:
            args.append(f"poissons_ratio={self.poissons_ratio!r}")
        if self.youngs_modulus is not None:
            args.append(f"youngs_modulus={self.youngs_modulus!r}")
        if self.shear_modulus is not None:
            args.append(f"shear_modulus={self.shear_modulus!r}")
        return ", ".join(args) + ", " + super().extra_repr()


class TotalVariation(_SpatialDerivativesLoss):
    r"""Total variation of displacement field."""

    def forward(self, u: Tensor) -> Tensor:
        r"""Evaluate regularization loss for given transformation."""
        spacing = self._spacing(u.shape)
        return L.total_variation_loss(
            u,
            mode=self.mode,
            sigma=self.sigma,
            spacing=spacing,
            stride=self.stride,
            reduction=self.reduction,
        )


TV = TotalVariation



r"""Custom itertools functions."""

from itertools import repeat
from typing import Any, Iterable, Sequence, Union


def is_even_permutation(permutation: Sequence[int]) -> bool:
    r"""Checks if given permutation is even.

    Example:
        >>> is_even_permutation(range(10))
        True
        >>> is_even_permutation(range(10)[::-1])
        False

    """
    if len(permutation) == 1:
        return True
    transitions_count = 0
    for index, element in enumerate(permutation):
        for next_element in permutation[index + 1 :]:
            if element > next_element:
                transitions_count += 1
    return not (transitions_count % 2)


def repeat_last(arg: Union[Any, Sequence[Any]], length: int) -> Sequence[Any]:
    r"""Repeat last element in sequence to extend it to the specified length."""
    if not isinstance(arg, str) and not isinstance(arg, Sequence):
        arg = (arg,)
    if not arg:
        raise ValueError("repeat_last() 'arg' must have at least one value to repeat")
    if len(arg) > length:
        raise ValueError("repeat_last() 'arg' sequence length must be at most '{length}'")
    arg = tuple(arg) + (arg[-1],) * (length - len(arg))
    return arg


def zip_longest_repeat_last(*args: Iterable):
    iterators = [iter(it) for it in args]
    num_active = len(iterators)
    if not num_active:
        return
    prev = None
    while True:
        values = []
        for i, it in enumerate(iterators):
            try:
                value = next(it)
            except StopIteration:
                num_active -= 1
                if not num_active:
                    return
                value = prev[i]
                iterators[i] = repeat(value)
            values.append(value)
        prev = tuple(values)
        yield prev



r"""Basic building blocks and (sub-)networks of learned image registration models.

For spatial transformation models used in both classic non-learning based registration and
learned image registration, where transformation parameters are inferred by a neural network
from the input data instead of being optimized directly, see the :mod:`.spatial` library.

Commonly, the neural network model infers the transformation parameters from the input data.
These parameters are then used to evaluate and apply the spatial transformation. For this, the
``params`` attribute of parameteric transformations can be set to a neural network instance of
type ``torch.nn.Module``. The input data of the so parametrized spatial transformation is then
set as :meth:`.SpatialTransform.condition`, which constitutes the input of the neural network.

"""



r"""Auxiliary functions for reading and writing images using SimpleITK."""

from typing import Tuple

from torch import Tensor

from deepali.core.grid import Grid
from deepali.core.pathlib import PathUri
from deepali.utils.simpleitk.imageio import read_image as _read_image
from deepali.utils.simpleitk.imageio import write_image as _write_image
from deepali.utils.simpleitk.torch import image_from_tensor, tensor_from_image


def read_sitk_image(path: PathUri) -> Tuple[Tensor, Grid]:
    r"""Read any image file supported by SimpleITK."""
    image = _read_image(path)
    data = tensor_from_image(image)
    grid = Grid.from_sitk(image)
    return data, grid


def write_sitk_image(data: Tensor, grid: Grid, path: PathUri, compress: bool = True) -> None:
    r"""Write image file in any format supported by SimpleITK."""
    origin = grid.origin().tolist()
    spacing = grid.spacing().tolist()
    direction = grid.direction().flatten().tolist()
    image = image_from_tensor(data, origin=origin, spacing=spacing, direction=direction)
    _write_image(image, path, compress=compress)



r"""Mix-in classes for torch.nn.Module subclasses."""

from typing import Dict, Optional

import torch
from torch.nn import Module


class DeviceProperty(object):
    r"""Mixin for torch.nn.Module to provide 'device' property."""

    @property
    def device(self) -> torch.device:
        r"""Device of first found module parameter or buffer."""
        for param in self.parameters():
            if param is not None:
                return param.device
        for buffer in self.buffers():
            if buffer is not None:
                return buffer.device
        return torch.device("cpu")


class ReprWithCrossReferences(object):
    r"""Mixin of __repr__ for torch.nn.Module subclasses to include cross-references to reused modules."""

    def __repr__(self) -> str:
        return self._repr_impl()

    def _repr_impl(
        self,
        prefix: str = "",
        module: Optional[Module] = None,
        memo: Optional[Dict[Module, str]] = None,
    ) -> str:
        if module is None:
            module = self
        if memo is None:
            memo = {}
        # We treat the extra repr like the sub-module, one item per line
        extra_lines = []
        extra_repr = module.extra_repr()
        # empty string will be split into list ['']
        if extra_repr:
            extra_lines = extra_repr.split("\n")
        child_lines = []
        for key, child in module._modules.items():
            mod_str = self._repr_impl(prefix=prefix + key + ".", module=child, memo=memo)
            mod_str = _addindent(mod_str, 2)
            prev_key = memo.get(child)
            if prev_key:
                mod_str = f"{prev_key}(\n  {mod_str}\n)"
                mod_str = _addindent(mod_str, 2)
            child_lines.append(f"({key}): {mod_str}")
            memo[child] = prefix + key
        lines = extra_lines + child_lines

        main_str = module._get_name() + "("
        if lines:
            # simple one-liner info, which most builtin Modules will use
            if len(extra_lines) == 1 and not child_lines:
                main_str += extra_lines[0]
            else:
                main_str += "\n  " + "\n  ".join(lines) + "\n"
        main_str += ")"
        return main_str


def _addindent(s_: str, numSpaces: int) -> str:
    r"""Add indentation to multi-line string."""
    # Copied from https://github.com/pytorch/pytorch/blob/992d251c39b5eb45e0b898feac46b18a0a8c8e8f/torch/nn/modules/module.py#L29-L38
    s = s_.split("\n")
    if len(s) == 1:
        return s_
    first = s.pop(0)
    s = [(numSpaces * " ") + line for line in s]
    s = "\n".join(s)
    s = first + "\n" + s
    return s



import numpy as np

import torch
from torch import Tensor

from deepali.core import tensor as U


def test_move_dim():
    # Same dimension
    for dim in range(5):
        shape = U.move_dim(torch.empty(1, 5, 7, 9, 3), dim, dim).shape
        assert shape == (1, 5, 7, 9, 3)
    # Move channels to last dimension
    shape = U.move_dim(torch.empty(1, 3, 5, 7, 9), 1, -1).shape
    assert shape == (1, 5, 7, 9, 3)
    shape = U.move_dim(torch.empty(1, 3, 5, 7, 9), 1, 4).shape
    assert shape == (1, 5, 7, 9, 3)
    # Move last dimension to channels dimension
    shape = U.move_dim(torch.empty(1, 5, 7, 9, 3), -1, 1).shape
    assert shape == (1, 3, 5, 7, 9)
    shape = U.move_dim(torch.empty(1, 5, 7, 9, 3), 4, 1).shape
    assert shape == (1, 3, 5, 7, 9)
    # Move arbitrary dimensions
    shape = U.move_dim(torch.empty(1, 5, 7, 9, 3), 0, 1).shape
    assert shape == (5, 1, 7, 9, 3)
    shape = U.move_dim(torch.empty(1, 5, 7, 9, 3), 1, 0).shape
    assert shape == (5, 1, 7, 9, 3)
    shape = U.move_dim(torch.empty(1, 5, 7, 9, 3), 0, 2).shape
    assert shape == (5, 7, 1, 9, 3)
    shape = U.move_dim(torch.empty(1, 5, 7, 9, 3), 2, 0).shape
    assert shape == (7, 1, 5, 9, 3)
    shape = U.move_dim(torch.empty(1, 5, 7, 9, 3), 2, 4).shape
    assert shape == (1, 5, 9, 3, 7)
    shape = U.move_dim(torch.empty(1, 5, 7, 9, 3), 4, 2).shape
    assert shape == (1, 5, 3, 7, 9)
    # Reversibility
    shape = U.move_dim(U.move_dim(torch.empty(1, 5, 7, 9, 3), 0, 2), 2, 0).shape
    assert shape == (1, 5, 7, 9, 3)


def test_unravel_coords():
    indices = torch.tensor([0, 1, 2, 3, 4, 5, 6, 7, 8], dtype=torch.int64)
    size = (3, 3)
    shape = tuple(reversed(size))
    coords = U.unravel_coords(indices, size)
    assert isinstance(coords, Tensor)
    assert coords.dtype == torch.int64

    expected = np.array(np.unravel_index(indices.numpy(), shape)).T
    expected = np.flip(expected, axis=-1)
    assert np.all(coords.numpy() == expected)

    indices = torch.tensor([0, 1, 2, 3, 4, 5, 6, 7, 8, 9], dtype=torch.int32)
    size = (2, 5)
    shape = tuple(reversed(size))
    coords = U.unravel_coords(indices, size)
    assert isinstance(coords, Tensor)
    assert coords.dtype == torch.int32

    expected = np.array(np.unravel_index(indices.numpy(), shape)).T
    expected = np.flip(expected, axis=-1)
    assert np.all(coords.numpy() == expected)


def test_unravel_index():
    indices = torch.tensor([0, 1, 2, 3, 4, 5, 6, 7, 8], dtype=torch.int64)
    shape = (3, 3)
    result = U.unravel_index(indices, shape)
    assert isinstance(result, Tensor)
    assert result.dtype == torch.int64

    expected = np.array(np.unravel_index(indices.numpy(), shape)).T
    assert np.all(result.numpy() == expected)

    indices = torch.tensor([0, 1, 2, 3, 4, 5, 6, 7, 8, 9], dtype=torch.int32)
    shape = (2, 5)
    result = U.unravel_index(indices, shape)
    assert isinstance(result, Tensor)
    assert result.dtype == torch.int32

    expected = np.array(np.unravel_index(indices.numpy(), shape)).T
    assert np.all(result.numpy() == expected)



r"""Auxiliary functions for working with `VTK <https://vtk.org/>`_ data structures."""

from .idlist import iter_id_list
from .idlist import iter_cell_point_ids
from .idlist import iter_point_cell_ids

from .polydataio import read_polydata
from .polydataio import write_polydata


# fmt: off
__all__ = (
    "iter_id_list",
    "iter_cell_point_ids",
    "iter_point_cell_ids",
    "read_polydata",
    "write_polydata",
)
# fmt: on



r"""Modules without learnable parameters.

This library defines subclasses of ``torch.nn.Module`` which expose the tensor operations
available in the :mod:`.core` library via a stateful functor object that can be used
in PyTorch models to perform predefined operations with in general no optimizable parameters.

"""

from .basic import GetItem
from .basic import Narrow
from .basic import Pad
from .basic import Reshape
from .basic import View

from .flow import Curl
from .flow import ExpFlow

from .image import BlurImage
from .image import FilterImage
from .image import GaussianConv

from .lambd import LambdaFunc
from .lambd import LambdaLayer

from .mixins import DeviceProperty
from .mixins import ReprWithCrossReferences

from .output import ToImmutableOutput

from .sample import AlignImage
from .sample import SampleImage
from .sample import TransformImage

from .utilities import remove_layers_in_state_dict
from .utilities import rename_layers_in_state_dict


__all__ = (
    "AlignImage",
    "BlurImage",
    "Curl",
    "DeviceProperty",
    "ExpFlow",
    "FilterImage",
    "GaussianConv",
    "GetItem",
    "LambdaFunc",
    "LambdaLayer",
    "Narrow",
    "Pad",
    "ReprWithCrossReferences",
    "Reshape",
    "SampleImage",
    "ToImmutableOutput",
    "TransformImage",
    "View",
    "remove_layers_in_state_dict",
    "rename_layers_in_state_dict",
)



r"""Event handlers for use with Ignite."""

from logging import Logger
from typing import Any, Callable, Dict, Mapping, Optional, Sequence, Set, Union

from ignite.engine import Engine, Events, State
from torch import Tensor
from torch.nn.modules import Module
from torch.optim.optimizer import Optimizer
from torch.utils.data import BatchSampler, DataLoader
from torch.utils.tensorboard import SummaryWriter

from deepali.core.collections import re_output_key_index
from deepali.utils.tensorboard import escape_channel_index_format_string, add_summary_images


def clamp_learning_rate(
    engine: Engine,
    optimizer: Optimizer,
    min_learning_rate: Optional[float] = None,
    max_learning_rate: Optional[float] = None,
):
    if min_learning_rate is None:
        min_learning_rate = 1e-12
    if max_learning_rate is None:
        max_learning_rate = float("inf")
    for param_group in optimizer.param_groups:
        param_group["lr"] = max(min_learning_rate, min(param_group["lr"], max_learning_rate))


def set_distributed_sampler_epoch(engine: Engine, epoch: Optional[int] = None) -> None:
    data = engine.state.dataloader
    if epoch is None:
        # -1 because this handler is usually invoked at start of epoch
        epoch = engine.state.epoch - 1
    if isinstance(data, DataLoader):
        set_sampler_epoch = getattr(data.sampler, "set_epoch", None)
        if callable(set_sampler_epoch):
            set_sampler_epoch(epoch)
        set_sampler_epoch = getattr(data.batch_sampler, "set_epoch", None)
        if callable(set_sampler_epoch):
            set_sampler_epoch(epoch)
        if isinstance(data.batch_sampler, BatchSampler):
            set_sampler_epoch = getattr(data.batch_sampler.sampler, "set_epoch", None)
            if callable(set_sampler_epoch):
                set_sampler_epoch(epoch)
        data = data.dataset
    set_epoch = getattr(data, "set_epoch", None)
    if callable(set_epoch):
        set_epoch(epoch)


def print_metrics(
    engine: Engine,
    prefix: Optional[str] = None,
    names: Optional[Union[Mapping[str, str], Sequence[str], Set[str]]] = None,
    logger: Optional[Logger] = None,
    global_step_transform: Optional[Callable] = None,
) -> None:
    r"""Log evaluated performance metrics stored in ``engine.state.metrics``."""
    metrics = get_scalar_metrics(engine.state)
    if not metrics:
        return
    if names is not None:
        metrics = {name: metrics[name] for name in names if name in metrics}
        if isinstance(names, Mapping):
            metrics = {names[k]: v for k, v in metrics.items()}
    global_step = get_global_step(engine, global_step_transform)
    if prefix is None:
        prefix = ""
        if global_step_transform is None:
            prefix = "epoch={e:03d}, "
        prefix += "iter={i:06d}, "
    msg = prefix.format(e=engine.state.epoch, i=global_step)
    msg += ", ".join(
        (
            f"{key}=" + (f"{value:.5f}" if isinstance(value, float) else f"{value}")
            for key, value in metrics.items()
        )
    )
    print(msg) if logger is None else logger.info(msg)


def reset_iterable_dataset(engine: Engine) -> None:
    r"""Reset iterator over engine.state.dataloader.

    This handler also calls ``set_distributed_sampler_epoch(engine)`` such that the
    epoch number is set prior to calling ``engine.set_data(engine.state.dataloader)``,
    which invokes ``iter(engine.state.dataloader)``. This is to ensure that if the
    latter shuffles the data using the epoch number as seed, that the dataset is
    then shuffled with the new epoch number.

    """
    set_distributed_sampler_epoch(engine, engine.state.epoch)
    engine.set_data(engine.state.dataloader)


def reset_model_grads(engine: Engine, model: Module) -> None:
    r"""Set ``grad`` attributes of model parameters to ``None``."""
    for p in model.parameters():
        p.grad = None


def reset_optimizer_grads(engine: Engine, optimizer: Optimizer) -> None:
    r"""Set ``grad`` attributes of optimizer parameters to ``None``."""
    for group in optimizer.param_groups:
        for p in group["params"]:
            p.grad = None


def set_engine_state(engine: Engine, **kwargs) -> None:
    r"""Add dictionary values as attributes to ``engine.state`` object."""
    for name, value in kwargs.items():
        setattr(engine.state, name, value)


def terminate_on_max_iteration(engine: Engine, max_iterations: int):
    r"""Terminate training when maximum number of global iterations reached."""
    if max_iterations > 0 and engine.state.iteration >= max_iterations:
        engine.terminate()


def write_summary_hists(
    engine: Engine,
    writer: SummaryWriter,
    model: Module,
    prefix: Optional[str] = None,
    weights: bool = True,
    grads: bool = True,
    global_step_transform: Optional[Callable] = None,
) -> None:
    r"""Add histograms of model weights to TensorBoard summary."""
    global_step = get_global_step(engine, global_step_transform)
    prefix = (prefix or "").format(e=engine.state.epoch, i=engine.state.iteration)
    for name, p in model.named_parameters():
        if p.grad is None:
            continue
        name = name.replace(".", "/")
        if weights:
            writer.add_histogram(
                tag=f"{prefix}weights/{name}",
                values=p.data.detach().cpu().numpy(),
                global_step=global_step,
            )
        if grads:
            writer.add_histogram(
                tag=f"{prefix}grads/{name}",
                values=p.grad.detach().cpu().numpy(),
                global_step=global_step,
            )


def write_summary_images(
    engine: Engine,
    writer: SummaryWriter,
    names: Optional[Union[Mapping[str, str], Sequence[str], Set[str]]] = None,
    prefix: Optional[str] = None,
    image_transform: Optional[Callable[[str, Tensor], Tensor]] = True,
    rescale_transform: Union[bool, Callable[[str, Tensor], Tensor]] = False,
    global_step_transform: Optional[Callable[[Engine, Events], int]] = None,
    channel_offset: Union[bool, Callable[[Engine], int]] = False,
) -> None:
    r"""Add images stored in ``engine.state.batch`` and ``engine.state.output`` to TensorBoard summary.

    Args:
        engine: Ignite engine.
        writer: TensorBoard writer with open event file.
        names: Possibly hierarchical keys of input batch or output image entries to write.
            When a dictionary is given, replace image name by the map value.
        prefix: Prefix string for TensorBoard tags. Can be format string with
            placeholders ``{e}`` for epoch and ``{i}`` global iteration.
        image_transform: Callable used to extract a 2D image tensor of shape
            ``(C, Y, X)`` from each image. When a multi-channel tensor is returnd (C > 1),
            each channel is saved as separate image to the TensorBoard event file.
            By default, the central slice of the first image in the batch is extracted.
            The first argument is the name of the image tensor if applicable, or an empty
            string otherwise. This can be used to differently process different images.
        rescale_transform: Image intensities must be in the closed interval ``[0, 1]``.
            Set to ``False``, if this is already the case or when a custom
            ``image_transform`` is used which normalizes the image intensities.
        global_step_transform: Callable used to obtain global iteration.
            Called with arguments ``engine`` and ``Events.ITERATION_COMPLETED``.
            If not specified, use ``engine.state.iteration``.
        channel_offset: Offset to use for channel index in image tag format string.
            If ``True``, use batch index times batch size of data loader plus one. If ``False``,
            use zero. Otherwise, a callable must be given which receives the engine on which
            the event is triggered as argument.

    """

    def filter_images(
        arg: Union[Mapping[str, Any], Sequence[Any]], prefix: str = ""
    ) -> Dict[str, Tensor]:
        images = {}
        if not isinstance(arg, dict):
            arg = {str(i): value for i, value in enumerate(arg)}
        for name, value in arg.items():
            if name == "loss":
                continue
            if isinstance(value, (dict, tuple, list)):
                images.update(filter_images(value, prefix + name + "."))
            elif not isinstance(value, Tensor) or value.ndim < 4:
                continue
            else:
                name = prefix + name
                if isinstance(names, dict):
                    if name not in names:
                        continue
                    name = names[name]
                elif names is not None and name not in names:
                    continue
                images[name] = value
        return images

    def format_tag(arg: str) -> str:
        arg = escape_channel_index_format_string(arg)
        arg = arg.format(e=engine.state.epoch, i=engine.state.iteration, b=batch_index)
        return arg

    batch_index = (engine.state.iteration - 1) % engine.state.epoch_length + 1
    global_step = get_global_step(engine, global_step_transform)
    channel_offset_value = 0
    if channel_offset is True:
        dataloader: DataLoader = engine.state.dataloader
        if not dataloader.batch_size:
            raise RuntimeError(
                "write_summary_images() 'channel_offset' is True, but 'engine.state.dataloader.batch_size' is None"
            )
        channel_offset_value = (batch_index - 1) * dataloader.batch_size + 1
    elif callable(channel_offset):
        channel_offset_value = channel_offset(engine)
    elif channel_offset is not False:
        raise TypeError("write_summary_images() 'channel_offset' must be bool or callable")
    # Format tag strings
    prefix = format_tag(prefix or "")
    if isinstance(names, dict):
        names = {re_output_key_index.sub(r".\1", name): tag for name, tag in names.items()}
        names = {name: format_tag(tag) for name, tag in names.items()}
    elif names is not None:
        names = {re_output_key_index.sub(r".\1", name) for name in names}
        names = {format_tag(name) for name in names}
    # Input batch
    batch = engine.state.batch
    if isinstance(batch, Tensor):
        batch = {"batch": batch}
    batch = filter_images(batch)
    add_summary_images(
        writer,
        prefix,
        batch,
        global_step=global_step,
        image_transform=image_transform,
        rescale_transform=rescale_transform,
        channel_offset=channel_offset_value,
    )
    # Model output
    output = engine.state.output
    if isinstance(output, Tensor):
        output = {"output": output}
    output = filter_images(output)
    add_summary_images(
        writer,
        prefix,
        output,
        global_step=global_step,
        image_transform=image_transform,
        rescale_transform=rescale_transform,
        channel_offset=channel_offset_value,
    )
    writer.flush()


def write_summary_metrics(
    engine: Engine,
    writer: SummaryWriter,
    prefix: Optional[str] = None,
    global_step_transform: Optional[Callable] = None,
):
    r"""Add computed values stored in ``engine.state.metrics`` to Tensorboard summary."""
    global_step = get_global_step(engine, global_step_transform)
    prefix = (prefix or "").format(e=engine.state.epoch, i=engine.state.iteration)
    metrics = get_scalar_metrics(engine.state)
    for name, value in metrics.items():
        writer.add_scalar(prefix + name, value, global_step=global_step)
    writer.flush()


def write_summary_optim_params(
    engine: Engine,
    writer: SummaryWriter,
    optimizer: Optimizer,
    params: Optional[Union[str, Sequence[str]]] = None,
    prefix: Optional[str] = None,
    global_step_transform: Optional[Callable] = None,
) -> None:
    r"""Add optimization parameters to TensorBoard summary."""
    global_step = get_global_step(engine, global_step_transform)
    prefix = (prefix or "").format(e=engine.state.epoch, i=engine.state.iteration)
    if isinstance(params, str):
        params = [params]
    for group_id, param_group in enumerate(optimizer.param_groups):
        for param_name in params or param_group.keys():
            try:
                param_value = float(param_group[param_name])
            except (KeyError, TypeError):
                continue
            tag = prefix + param_name
            if len(optimizer.param_groups) > 1:
                tag += f"/group_{group_id}"
            writer.add_scalar(tag, param_value, global_step)


def get_global_step(engine: Engine, global_step_transform: Optional[Callable] = None) -> int:
    r"""Get global step for summary event."""
    if global_step_transform is None:
        return engine.state.iteration
    global_step = global_step_transform(engine, Events.ITERATION_COMPLETED)
    if not isinstance(global_step, int):
        raise TypeError(f"'global_step_transform' must return an int, got {type(global_step)}")
    return global_step


def get_scalar_metrics(state: State) -> Dict[str, Union[float, int]]:
    r"""Get scalar metrics stored in ``state.metrics``."""
    metrics = {}
    for key, value in state.metrics.items():
        if isinstance(value, Tensor):
            if value.squeeze().dim() != 0:
                continue
            value = value.item()
            value = int(value) if isinstance(value, int) else float(value)
        if isinstance(value, (float, int)):
            metrics[key] = value
    return metrics



r"""Non-rigid transformation models."""

from __future__ import annotations

from copy import copy as shallow_copy
import math
from typing import Callable, Optional, Sequence, TypeVar, Union, cast

import torch
from torch import Size, Tensor

from deepali.core import functional as U
from deepali.core.grid import Axes, Grid
from deepali.data.flow import FlowFields
from deepali.modules import ExpFlow

from .base import NonRigidTransform
from .parametric import ParametricTransform


TDenseVectorFieldTransform = TypeVar(
    "TDenseVectorFieldTransform", bound="DenseVectorFieldTransform"
)


class DenseVectorFieldTransform(ParametricTransform, NonRigidTransform):
    r"""Dense vector field transformation with linear interpolation at non-grid point locations."""

    def __init__(
        self,
        grid: Grid,
        groups: Optional[int] = None,
        params: Optional[Union[bool, Tensor, Callable[..., Tensor]]] = True,
        stride: Optional[Union[float, Sequence[float]]] = None,
        resize: bool = True,
    ) -> None:
        r"""Initialize transformation parameters.

        Args:
            grid: Grid domain on which transformation is defined.
            groups: Number of transformations. A given image batch can either be deformed by a
                single transformation, or a separate transformation for each image in the batch, e.g.,
                for group-wise or batched registration. The default is one transformation for all images
                in the batch, or the batch length of the ``params`` tensor if provided.
            params: Initial parameters. If a tensor is given, it is only registered as optimizable module
                parameters when of type ``torch.nn.Parameter``. When a callable is given instead, it will be
                called by ``self.update()`` with arguments set and given by ``self.condition()``. When a boolean
                argument is given, a new zero-initialized tensor is created. If ``True``, this tensor is registered
                as optimizable module parameter.
            stride: Spacing between vector field grid points in units of input ``grid`` points.
                Can be used to subsample the dense vector field with respect to the image grid of the fixed target
                image. When ``grid.align_corners() is True``, the corner points of the ``grid`` and the resampled
                vector field grid are aligned. Otherwise, the edges of the grid domains are aligned.
            resize: Whether to resize vector field during transformation update. If ``True``, the buffered vector
                field ``u`` (and ``v`` if applicable) is resized to match the image ``grid`` size. This means that
                transformation constraints defined on these resized vector fields, such as those based on finite
                differences, are evaluated at the image grid resolution rather than the resolution of the underlying
                vector field parameterization. This influences the scale at which these constraints are imposed.

        """
        if stride is None:
            stride = 1
        if isinstance(stride, (int, float)):
            stride = (stride,) * grid.ndim
        if len(stride) != grid.ndim:
            raise ValueError(
                f"{type(self).__name__}() 'stride' must be float or Sequence of length {grid.ndim}"
            )
        self.stride = tuple(float(s) for s in stride)
        self._resize = resize
        super().__init__(grid, groups=groups, params=params)

    @property
    def data_shape(self) -> Size:
        r"""Get shape of transformation parameters tensor."""
        grid = self.grid()
        shape = self.data_grid_shape(grid)
        return Size((grid.ndim,) + shape)

    def data_grid(
        self, grid: Optional[Grid] = None, stride: Optional[Union[float, Sequence[float]]] = None
    ) -> Grid:
        if grid is None:
            grid = self.grid()
        if stride is None:
            stride = self.stride
        return grid.reshape(self.data_grid_shape(grid, stride))

    def data_grid_shape(
        self, grid: Optional[Grid] = None, stride: Optional[Union[float, Sequence[float]]] = None
    ) -> torch.Size:
        if grid is None:
            grid = self.grid()
        if stride is None:
            stride = self.stride
        return tuple(int(math.ceil(n / s)) for n, s in zip(grid.shape, stride))

    @torch.no_grad()
    def grid_(self: TDenseVectorFieldTransform, grid: Grid) -> TDenseVectorFieldTransform:
        r"""Set sampling grid of transformation domain and codomain.

        If ``self.params`` is a callable, only the grid attribute is updated, and
        the callable must return a tensor of matching size upon next evaluation.

        """
        params = self.params
        if isinstance(params, Tensor):
            prev_grid = self._grid
            grid_axes = Axes.from_grid(grid)
            flow_axes = self.axes()
            flow_grid = prev_grid.reshape(params.shape[2:])
            flow = FlowFields(params, grid=flow_grid, axes=flow_axes)
            flow = flow.sample(self.data_grid(grid))
            flow = flow.axes(grid_axes)
            # Change self._grid before self.data_() as it defines self.data_shape
            super().grid_(grid)
            try:
                self.data_(flow.tensor())
            except Exception:
                self._grid = prev_grid
                raise
        else:
            super().grid_(grid)
        return self

    def evaluate(self, resize: Optional[bool] = None) -> Tensor:
        r"""Update buffered displacement vector field."""
        if resize is None:
            resize = self._resize
        u = self.data()
        u = u.view(*u.shape)  # such that named_buffers() returns both 'u' (or 'v') and 'p'
        if resize:
            align_corners = self.align_corners()
            grid_shape = self.grid().shape
            u = U.grid_reshape(u, grid_shape, align_corners=align_corners)
        return u


class DisplacementFieldTransform(DenseVectorFieldTransform):
    r"""Dense displacement field transformation model."""

    def fit(self, flow: FlowFields, **kwargs) -> DisplacementFieldTransform:
        r"""Fit transformation to a given flow field.

        Args:
            flow: Flow fields to approximate.
            kwargs: Optional keyword arguments are ignored.

        Returns:
            Reference to this transformation.

        Raises:
            RuntimeError: When this transformation has no optimizable parameters.

        """
        params = self.params
        if params is None:
            raise AssertionError(f"{type(self).__name__}.data() 'params' must be set first")
        grid = self.grid()
        if not callable(params):
            grid = self.grid().resize(self.data_shape[:1:-1])
        flow = flow.to(self.device)
        flow = flow.sample(grid)
        flow = flow.axes(grid.axes())
        if callable(params):
            self._fit(flow, **kwargs)
        else:
            self.data_(flow.tensor())
        return self

    def update(self) -> DisplacementFieldTransform:
        r"""Update buffered displacement vector field."""
        super().update()
        u = self.evaluate()
        self.register_buffer("u", u, persistent=False)
        return self


class StationaryVelocityFieldTransform(DenseVectorFieldTransform):
    r"""Dense stationary velocity field transformation."""

    def __init__(
        self,
        grid: Grid,
        groups: Optional[int] = None,
        params: Optional[Union[bool, Tensor, Callable[..., Tensor]]] = True,
        stride: Optional[Union[float, Sequence[float]]] = None,
        resize: bool = True,
        scale: Optional[float] = None,
        steps: Optional[int] = None,
    ) -> None:
        r"""Initialize transformation parameters.

        Args:
            grid: Grid on which to sample velocity field vectors.
            groups: Number of velocity fields. A given image batch can either be deformed by a
                single displacement field, or one separate displacement field for each image in the
                batch, e.g., for group-wise or batched registration. The default is one displacement
                field for all images in the batch, or the batch length N of ``params`` if provided.
            params: Initial parameters of velocity fields of shape ``(N, C, ...X)``, where N must match
                the value of ``groups``, and vector components are the image channels in the order x, y, z.
                Note that a tensor is only registered as optimizable module parameters when of type
                ``torch.nn.Parameter``. When a callable is given instead, it will be called each time the
                model parameters are accessed with the arguments set and returned by ``self.condition()``.
                When a boolean argument is given, a new zero-initialized tensor is created. If ``True``,
                it is registered as optimizable parameter.
            stride: Spacing between vector field grid points in units of input ``grid`` points.
                Can be used to subsample the dense vector field with respect to the image grid of the fixed target
                image. When ``grid.align_corners() is True``, the corner points of the ``grid`` and the resampled
                vector field grid are aligned. Otherwise, the edges of the grid domains are aligned.
            resize: Whether to resize vector field during transformation update. If ``True``, the buffered vector
                fields ``v`` and ``u`` are resized to match the image ``grid`` size. This means that transformation
                constraints defined on these resized vector fields, such as those based on finite differences, are
                evaluated at the image grid resolution rather than the resolution of the underlying vector field
                parameterization. This influences the scale at which these constraints are imposed.
            scale: Constant scaling factor of velocity fields.
            steps: Number of scaling and squaring steps.

        """
        super().__init__(grid, groups=groups, params=params, stride=stride, resize=resize)
        self.exp = ExpFlow(scale=scale, steps=steps, align_corners=grid.align_corners())

    def grid_(self, grid: Grid) -> StationaryVelocityFieldTransform:
        r"""Set sampling grid of transformation domain and codomain."""
        super().grid_(grid)
        self.exp.align_corners = grid.align_corners()
        return self

    def inverse(
        self, link: bool = False, update_buffers: bool = False
    ) -> StationaryVelocityFieldTransform:
        r"""Get inverse of this transformation.

        Args:
            link: Whether to inverse transformation keeps a reference to this transformation.
                If ``True``, the ``update()`` function of the inverse function will not recompute
                shared parameters, e.g., parameters obtained by a callable neural network, but
                directly access the parameters from this transformation. Note that when ``False``,
                the inverse transformation will still share parameters, modules, and buffers with
                this transformation, but these shared tensors may be replaced by a call of ``update()``
                (which is implicitly called as pre-forward hook when ``__call__()`` is invoked).
            update_buffers: Whether buffers of inverse transformation should be update after creating
                the shallow copy. If ``False``, the ``update()`` function of the returned inverse
                transformation has to be called before it is used.

        Returns:
            Shallow copy of this transformation with ``exp`` module which uses negative scaling factor
            to scale and square the stationary velocity field to computes the inverse displacement field.

        """
        inv = shallow_copy(self)
        if link:
            inv.link_(self)
        inv.exp = cast(ExpFlow, self.exp).inverse()
        if update_buffers:
            v = getattr(inv, "v", None)
            if v is not None:
                u = inv.exp(v)
                inv.register_buffer("u", u, persistent=False)
        return inv

    def update(self) -> StationaryVelocityFieldTransform:
        r"""Update buffered velocity and displacement vector fields."""
        super().update()
        v = self.evaluate()
        u = self.exp(v)
        self.register_buffer("v", v, persistent=False)
        self.register_buffer("u", u, persistent=False)
        return self



r"""Non-linear activation functions."""

from functools import partial
from typing import Any, Callable, Mapping, Optional, Sequence, Type, Union

from torch import Tensor, nn
from torch.nn import Module

from .lambd import LambdaLayer


ActivationFunc = Callable[[Tensor], Tensor]
ActivationArg = Union[ActivationFunc, str, Mapping, Sequence, None]


ACTIVATION_TYPES = {
    "celu": nn.CELU,
    "elu": nn.ELU,
    "hardtanh": nn.Hardtanh,
    "identity": nn.Identity,
    "none": nn.Identity,
    "relu": nn.ReLU,
    "relu6": nn.ReLU6,
    "lrelu": nn.LeakyReLU,
    "leakyrelu": nn.LeakyReLU,
    "leaky_relu": nn.LeakyReLU,
    "rrelu": nn.RReLU,
    "selu": nn.SELU,
    "gelu": nn.GELU,
    "hardshrink": nn.Hardshrink,
    "hardsigmoid": nn.Hardsigmoid,
    "hardswish": nn.Hardswish,
    "logsigmoid": nn.LogSigmoid,
    "logsoftmax": nn.LogSoftmax,
    "prelu": nn.PReLU,
    "sigmoid": nn.Sigmoid,
    "softmax": nn.Softmax,
    "softmin": nn.Softmin,
    "softplus": nn.Softplus,
    "softshrink": nn.Softshrink,
    "softsign": nn.Softsign,
    "tanh": nn.Tanh,
    "tanhshrink": nn.Tanhshrink,
}

INPLACE_ACTIVATIONS = {
    "elu",
    "hardtanh",
    "lrelu",
    "leakyrelu",
    "relu",
    "relu6",
    "rrelu",
    "selu",
    "celu",
}

SOFTMINMAX_ACTIVATIONS = {"softmin", "softmax", "logsoftmax"}


def activation(
    arg: ActivationArg,
    *args: Any,
    dim: Optional[int] = None,
    inplace: Optional[bool] = None,
    **kwargs,
) -> Module:
    r"""Get activation function.

    Args:
        arg: Custom activation function or module, or name of activation function with optional keyword arguments.
        args: Arguments to pass to activation init function.
        dim: Dimension along which to compute softmax activations (cf. ``ACT_SOFTMINMAX``). Unused by other activations.
        inplace: Whether to compute activation output in place. Unused if unsupported by specified activation function.
        kwargs: Additional keyword arguments for activation function. Overrides keyword arguments given as second
            tuple item when ``arg`` is a ``(name, kwargs)`` tuple instead of a string.

    Returns:
        Given activation function when ``arg`` is a ``torch.nn.Module``, or a new activation module otherwise.

    """
    if isinstance(arg, Module) and not args and not kwargs:
        return arg
    if callable(arg):
        return Activation(arg, *args, **kwargs)
    acti_name = "identity"
    acti_args = {}
    if isinstance(arg, str):
        acti_name = arg
    elif isinstance(arg, Mapping):
        acti_name = arg.get("name")
        if not acti_name:
            raise ValueError("activation() 'arg' map must contain 'name'")
        if not isinstance(acti_name, str):
            raise TypeError("activation() 'name' must be str")
        acti_args = {key: value for key, value in arg.items() if key != "name"}
    elif isinstance(arg, Sequence):
        if len(arg) != 2:
            raise ValueError("activation() 'arg' sequence must have length two")
        acti_name, acti_args = arg
        if not isinstance(acti_name, str):
            raise TypeError("activation() first 'arg' sequence argument must be str")
        if not isinstance(acti_args, dict):
            raise TypeError("activation() second 'arg' sequence argument must be dict")
        acti_args = acti_args.copy()
    elif arg is not None:
        raise TypeError("activation() 'arg' must be str, mapping, 2-tuple, callable, or None")
    acti_name = acti_name.lower()
    acti_type: Type[Module] = ACTIVATION_TYPES.get(acti_name)
    if acti_type is None:
        raise ValueError(
            f"activation() 'arg' name {acti_name!r} is unknown."
            " Pass a callable activation function or module instead."
        )
    acti_args.update(kwargs)
    if inplace is not None and acti_name in INPLACE_ACTIVATIONS:
        acti_args["inplace"] = bool(inplace)
    if acti_name in SOFTMINMAX_ACTIVATIONS:
        if dim is None and len(args) == 1 and isinstance(args, int):
            dim = args[0]
        elif args or acti_args:
            raise ValueError("activation() named {act_name!r} has no parameters")
        if dim is None:
            dim = 1
        acti = acti_type(dim)
    else:
        acti = acti_type(*args, **acti_args)
    return acti


class Activation(LambdaLayer):
    r"""Non-linear activation function."""

    def __init__(
        self,
        arg: ActivationArg,
        *args: Any,
        dim: int = 1,
        inplace: Optional[bool] = None,
        **kwargs,
    ) -> None:
        if callable(arg):
            acti = partial(arg, *args, **kwargs) if args or kwargs else arg
        else:
            acti = activation(arg, *args, dim=dim, inplace=inplace, **kwargs)
        super().__init__(acti)


def is_activation(arg: Any) -> bool:
    r"""Whether given object is an non-linear activation function module."""
    if isinstance(arg, Activation):
        return True
    types = tuple(ACTIVATION_TYPES.values())
    return isinstance(arg, types)



r"""Modules operating on flow vector fields."""

from __future__ import annotations

from copy import copy as shallow_copy
from typing import Optional, Union

from torch import Tensor
from torch.nn import Module

from deepali.core import ALIGN_CORNERS
from deepali.core import Array, Scalar, ScalarOrTuple
from deepali.core import functional as U


class Curl(Module):
    r"""Layer which calculates the curl of a vector field."""

    def __init__(
        self,
        mode: Optional[str] = None,
        sigma: Optional[float] = None,
        spacing: Optional[Union[Scalar, Array]] = None,
        stride: Optional[ScalarOrTuple[int]] = None,
    ) -> None:
        super().__init__()
        self.mode = mode
        self.sigma = sigma
        self.spacing = spacing
        self.stride = stride

    def forward(self, x: Tensor) -> Tensor:
        return U.curl(x, mode=self.mode, sigma=self.sigma, spacing=self.spacing, stride=self.stride)

    def extra_repr(self) -> str:
        args = []
        if self.mode is not None:
            args.append(f"mode={self.mode!r}")
        if self.sigma is not None:
            args.append(f"sigma={self.sigma!r}")
        if self.spacing is not None:
            args.append(f"spacing={self.spacing!r}")
        if self.stride is not None:
            args.append(f"stride={self.stride!r}")
        return ", ".join(args)


class ExpFlow(Module):
    r"""Layer that computes exponential map of flow field."""

    def __init__(
        self,
        scale: Optional[float] = None,
        steps: Optional[int] = None,
        align_corners: bool = ALIGN_CORNERS,
    ):
        r"""Initialize parameters.

        Args:
            scale: Constant scaling factor of input velocities (e.g., -1 for inverse). Default is 1.
            steps: Number of squaring steps.
            align_corners: Whether input vectors are with respect to ``Axes.CUBE`` (False)
                or ``Axes.CUBE_CORNERS`` (True). This flag is passed on to ``grid_sample()``.

        """
        super().__init__()
        self.scale = float(1 if scale is None else scale)
        self.steps = int(5 if steps is None else steps)
        self.align_corners = bool(align_corners)

    def forward(self, x: Tensor, inverse: bool = False) -> Tensor:
        r"""Compute exponential map of vector field."""
        scale = self.scale
        if inverse:
            scale *= -1
        return U.expv(x, scale=scale, steps=self.steps, align_corners=self.align_corners)

    @property
    def inv(self) -> ExpFlow:
        r"""Get inverse exponential map.

        .. code-block:: python

            u = exp(v)
            w = exp.inv(v)

        """
        return self.inverse()

    def inverse(self) -> ExpFlow:
        r"""Get inverse exponential map."""
        copy = shallow_copy(self)
        copy.scale *= -1
        return copy

    def extra_repr(self) -> str:
        return f"scale={repr(self.scale)}, steps={repr(self.steps)}"



r"""Oriented regularly spaced data sampling grid."""

from __future__ import annotations

from copy import copy as shallow_copy
from enum import Enum
from typing import TYPE_CHECKING, Any, Dict, Optional, Sequence, Union, overload

import numpy as np
from pkg_resources import parse_version

try:
    import SimpleITK as _sitk
except ImportError:
    _sitk = None

import torch
import torch.nn.functional as F
from torch import Tensor

if TYPE_CHECKING:
    # At runtime, cyclical import resolved by Grid.cube() instead
    from .cube import Cube

from .enum import SpatialDim, SpatialDimArg
from .math import round_decimals
from .linalg import homogeneous_transform, homogeneous_matrix, hmm
from .tensor import as_tensor, cat_scalars
from .typing import Array, Device, DType, PathStr, ScalarOrTuple, Shape, Size
from .typing import is_int_dtype


ALIGN_CORNERS = True
r"""By default, grid corner points define the domain within which the data is defined.

The normalized coordinates with respect to the grid cube are thus in [-1, 1] between
the first and last grid points (grid cell center points). This is such that when a
grid is downsampled or upsampled, respectively, the grid points at the boundary of
the domain remain unchanged. Points are only inserted or removed between these.

"""


class Axes(Enum):
    r"""Enumeration of grid axes with respect to which grid coordinates are defined."""

    GRID = "grid"
    r"""Oriented along grid axes with units corresponding to voxel units with origin
       with respect to world space at grid/image point with zero indices."""

    CUBE = "cube"
    r"""Oriented along grid axes with units corresponding to unit cube :math:`[-1, 1]^D`
       with origin with respect to world space at grid/image center and extrema -1/1
       coinciding with the grid border (``align_corners=False``)."""

    CUBE_CORNERS = "cube_corners"
    r"""Oriented along grid axes with units corresponding to unit cube :math:`[-1, 1]^D`
       with origin with respect to world space at grid/image center and extrema -1/1
       coinciding with the grid corner points (``align_corners=True``)."""

    WORLD = "world"
    r"""Oriented along world axes (physical space) with units corresponding to grid spacing (mm)."""

    @classmethod
    def from_arg(cls, arg: Union[Axes, str, None]) -> Axes:
        r"""Create enumeration value from function argument."""
        if arg is None or arg == "default":
            return cls.CUBE_CORNERS if ALIGN_CORNERS else cls.CUBE
        return cls(arg)

    @classmethod
    def from_grid(cls, grid: Grid) -> Axes:
        r"""Create enumeration value from sampling grid object."""
        return cls.from_align_corners(grid.align_corners())

    @classmethod
    def from_align_corners(cls, align_corners: bool) -> Axes:
        r"""Create enumeration value from sampling grid object."""
        return cls.CUBE_CORNERS if align_corners else cls.CUBE


class Grid(object):
    r"""Oriented regularly spaced data sampling grid.

    The dimensions of :attr:`.Grid.shape` are in reverse order of the dimensions of :meth:`.Grid.size`.
    The latter is consistent with SimpleITK, and the order of coordinates of :meth:`.Grid.origin`,
    :meth:`.Grid.spacing`, and :meth:`.Grid.direction`. Property :attr:`.Grid.shape`, on the other hand,
    is consistent with the order of dimensions of an image data tensor of type ``torch.Tensor``.

    To not confuse :meth:`.Grid.size` with ``torch.Tensor.size()``, it is recommended to prefer
    property ``torch.Tensor.shape``. The ``shape`` property is also known from ``numpy.ndarray.shape``.

    A :class:`.Grid` instance stores the grid center point instead of the origin corresponding to the grid
    point with zero index along each grid dimension. This simplifies resizing and resampling operations,
    which do not need to modify the origin explicitly, but keep the center point fixed. To get the
    coordinates of the grid origin, use :meth:`.Grid.origin`. For convenience, the :class:`.Grid`
    initialization function also accepts an ``origin`` instead of a ``center`` point as keyword argument.
    Conversion between center point and origin are taken care internally. When both ``origin`` and ``center``
    are specified, an error is raised if these are inconsistent with one another.

    In addition, :meth:`.Grid.points`, :meth:`.Grid.transform`, and :meth`.Grid.apply_transform` support
    coordinates with respect to different axes: 1) (continuous) grid indices, 2) world space, and
    3) grid-aligned cube with side length 2. The latter, i.e., :attr:`.Axes.CUBE` or :attr:`.Axes.CUBE_CORNERS`
    makes coordinates independent of :meth:`.Grid.size` and :meth:`.Grid.spacing`. These normalized coordinates
    are furthermore compatible with the ``grid`` argument of ``torch.nn.functional.grid_sample()``. Use
    :meth:`.Grid.cube` to obtain a :class:`.Cube` defining the data domain without spatial sampling attributes.

    """

    __slots__ = (
        "_size",
        "_center",
        "_spacing",
        "_direction",
        "_align_corners",
    )

    def __init__(
        self,
        size: Optional[Union[Size, Array]] = None,
        shape: Optional[Union[Shape, Array]] = None,
        center: Optional[Union[Array, float]] = None,
        origin: Optional[Union[Array, float]] = None,
        spacing: Optional[Union[Array, float]] = None,
        direction: Optional[Array] = None,
        device: Optional[Device] = None,
        align_corners: bool = ALIGN_CORNERS,
    ):
        r"""Initialize sampling grid attributes.

        Args:
            size: Size of spatial grid dimensions in the order ``(X, ...)``.
            shape: Size of spatial grid dimensions in the order ``(..., X)``.
                Must be ``None`` or ``size`` reversed, if ``size is not None``.
                Either ``size`` or ``shape`` must be specified.
            center: Grid center point in world space.
            origin: World coordinates of grid point with zero indices.
            spacing: Size of each grid square along each dimension.
            direction: Direction cosines defining orientation of grid in world space.
                The direction cosines are vectors that point from one pixel to the next.
                Each column of the matrix indicates the direction cosines of the unit vector
                that is parallel to the lines of the grid corresponding to that dimension.
            device: Device on which to store grid attributes. If ``None``, use ``"cpu"``.
            align_corners: Whether position of grid corner points are preserved by grid
                resampling and resizing operations by default. If ``True``, the grid
                origin remains unchanged by grid resizing operations, but the grid extent
                may in total change by one times the spacing between grid points. If ``False``,
                the extent of the grid remains constant, but the grid origin may shift.

        """
        if device is None:
            device = torch.device("cpu")
        # Grid size, optionally given as data tensor shape
        if size is not None:
            size = as_tensor(size, device=device)
            if size.ndim != 1:
                raise ValueError("Grid() 'size' must be 1-dimensional array")
            if len(size) == 0:
                raise ValueError("Grid() 'size' must be non-empty array")
        if shape is None:
            if size is None:
                raise ValueError("Grid() 'size' or 'shape' required")
        else:
            shape = as_tensor(shape, device=device)
            if size is None:
                if shape.ndim != 1:
                    raise ValueError("Grid() 'shape' must be 1-dimensional array")
                if len(shape) == 0:
                    raise ValueError("Grid() 'shape' must be non-empty array")
                size = shape.flip(0)
            else:
                with torch.no_grad():
                    if len(size) != len(shape) or shape.flip(0).ne(size).any():
                        raise ValueError("Grid() 'size' and 'shape' are not compatible")
        # Store size as float such that ``grid.downsample().upsample() == grid```
        self._size = torch.clamp(size.float(), min=0)
        # Set other properties AFTER _size, which defines 'device', 'dim' properties.
        # Use in-place setters to take care of conversion and value assertions.
        self.spacing_(1 if spacing is None else spacing)
        if direction is None:
            direction = torch.eye(self.ndim)
        self.direction_(direction)
        # Set grid center to default, specified center point, or derived from origin
        if origin is None:
            self.center_(0 if center is None else center)
        elif center is None:
            # ATTENTION: This must be done AFTER size, spacing, and direction are set!
            self.origin_(origin)
        else:
            self.center_(center)
            with torch.no_grad():
                origin = cat_scalars(origin, num=self.ndim, dtype=self.dtype, device=self.device)
                if not torch.allclose(origin, self.origin()):
                    raise ValueError("Grid() 'center' and 'origin' are inconsistent")
        # Default align_corners option argument for grid resizing operations
        self._align_corners = bool(align_corners)

    def numpy(self) -> np.ndarray:
        r"""Get grid attributes as 1-dimensional NumPy array."""
        return np.concatenate(
            [
                self._size.numpy(),
                self._spacing.numpy(),
                self._center.numpy(),
                self._direction.flatten().numpy(),
            ],
            axis=0,
        )

    @classmethod
    def from_numpy(
        cls,
        attrs: Union[Sequence[float], np.ndarray],
        origin: bool = False,
        align_corners: bool = ALIGN_CORNERS,
    ) -> Grid:
        r"""Create Grid from 1-dimensional NumPy array."""
        if isinstance(attrs, np.ndarray):
            seq = attrs.astype(float).tolist()
        else:
            seq = attrs
        return cls.from_seq(seq, origin=origin, align_corners=align_corners)

    @classmethod
    def from_seq(
        cls, attrs: Sequence[float], origin: bool = False, align_corners: bool = ALIGN_CORNERS
    ) -> Grid:
        r"""Create Grid from sequence of attribute values.

        Args:
            attrs: Array of length (D + 3) * D, where ``D=2`` or ``D=3`` is the number
                of spatial grid dimensions and array items are given as
                ``(nx, ..., sx, ..., cx, ..., d11, ..., d21, ....)``,
                where ``(nx, ...)`` is the grid size, ``(sx, ...)`` the grid spacing,
                ``(cx, ...)`` the grid center coordinates, and ``(d11, ...)``
                are the grid direction cosines. The argument can be a Python
                list or tuple, NumPy array, or PyTorch tensor.
            origin: Whether ``(cx, ...)`` specifies Grid origin rather than center.

        Returns:
            Grid instance.

        """
        if len(attrs) == 10:
            d = 2
        elif len(attrs) == 18:
            d = 3
        else:
            raise ValueError(
                f"{cls.__name__}.from_seq() expected array of length 10 (D=2) or 18 (D=3)"
            )
        kwargs = dict(
            size=attrs[0:d],
            spacing=attrs[d : 2 * d],
            direction=attrs[3 * d :],
            align_corners=align_corners,
        )
        if origin:
            kwargs["origin"] = attrs[2 * d : 3 * d]
        else:
            kwargs["center"] = attrs[2 * d : 3 * d]
        return Grid(**kwargs)

    @classmethod
    def from_batch(cls, tensor: Tensor) -> Grid:
        r"""Create default grid from (image) batch tensor.

        Args:
            tensor: Batch tensor of shape ``(N, C, ..., X)``.

        Returns:
            New default grid with size ``(X, ...)``.

        """
        return cls(shape=tensor.shape[2:])

    @classmethod
    def from_file(cls, path: PathStr, align_corners: bool = ALIGN_CORNERS) -> Grid:
        r"""Create sampling grid from image file header information."""
        if _sitk is None:
            raise RuntimeError(f"{cls.__name__}.from_file() requires SimpleITK")
        reader = _sitk.ImageFileReader()
        reader.SetFileName(str(path))
        reader.ReadImageInformation()
        return cls.from_reader(reader, align_corners=align_corners)

    @classmethod
    def from_reader(
        cls,
        reader: "_sitk.ImageFileReader",  # type: ignore
        align_corners: bool = ALIGN_CORNERS,
    ) -> Grid:
        r"""Create sampling grid from image file reader attributes."""
        return cls(
            size=reader.GetSize(),
            origin=reader.GetOrigin(),
            spacing=reader.GetSpacing(),
            direction=reader.GetDirection(),
            align_corners=align_corners,
        )

    @classmethod
    def from_sitk(
        cls,
        image: "_sitk.Image",  # type: ignore
        align_corners: bool = ALIGN_CORNERS,
    ) -> Grid:
        r"""Create sampling grid from ``SimpleITK.Image`` attributes."""
        return cls(
            size=image.GetSize(),
            origin=image.GetOrigin(),
            spacing=image.GetSpacing(),
            direction=image.GetDirection(),
            align_corners=align_corners,
        )

    def cube(self) -> "Cube":
        r"""Get oriented cube defining the space of normalized coordinates."""
        # Import locally due to cyclical dependency between Cube and Grid
        from .cube import Cube

        return Cube(
            extent=self.cube_extent(),
            center=self.center(),
            direction=self.direction(),
            device=self.device,
        )

    def domain(self) -> "Cube":
        r"""Get oriented bounding box defining the sampling grid domain in world space."""
        return self.cube()

    def dim(self) -> int:
        r"""Number of spatial grid dimensions."""
        return len(self._size)

    @property
    def ndim(self) -> int:
        r"""Number of grid dimensions."""
        return len(self._size)

    @property
    def dtype(self) -> torch.dtype:
        r"""Get data type of grid attribute tensors."""
        return self._size.dtype

    @property
    def device(self) -> torch.device:
        r"""Get device on which grid attribute tensors are stored."""
        return self._size.device

    def clone(self) -> Grid:
        r"""Make deep copy of this ``Grid`` instance."""
        grid = shallow_copy(self)
        for name in self.__slots__:
            value = getattr(self, name)
            if isinstance(value, Tensor):
                setattr(grid, name, value.clone())
        return grid

    def __deepcopy__(self, memo) -> Grid:
        r"""Support copy.deepcopy to clone this grid."""
        if id(self) in memo:
            return memo[id(self)]
        copy = self.clone()
        memo[id(self)] = copy
        return copy

    @overload
    def align_corners(self) -> bool:
        """Whether resizing operations preserve grid extent (False) or corner points (True)."""
        ...

    @overload
    def align_corners(self, arg: bool) -> Grid:
        """Set if resizing operations preserve grid extent (False) or corner points (True)."""
        ...

    def align_corners(self, arg: Optional[bool] = None) -> Union[bool, Grid]:
        """Whether resizing operations preserve grid extent (False) or corner points (True)."""
        if arg is None:
            return self._align_corners
        return shallow_copy(self).align_corners_(arg)

    def align_corners_(self, arg: bool) -> Grid:
        """Set if resizing operations preserve grid extent (False) or corner points (True)."""
        self._align_corners = bool(arg)
        return self

    def axes(self) -> Axes:
        r"""Grid axes."""
        return Axes.from_grid(self)

    def numel(self) -> int:
        r"""Number of grid points."""
        return self.size().numel()

    @staticmethod
    def _round_size(size: Tensor) -> Tensor:
        r"""Round sampling grid size attribute."""
        zero = torch.tensor(0, dtype=size.dtype, device=size.device)
        return torch.where(size.eq(zero), zero, size.ceil())

    def size_tensor(self) -> Tensor:
        r"""Sampling grid size as floating point tensor."""
        return self._round_size(self._size)

    @overload
    def size(self, i: int) -> int:
        r"""Sampleing grid size along the specified spatial dimension."""
        ...

    @overload
    def size(self) -> torch.Size:
        r"""Sampling grid size for dimensions ordered as ``(X, ...)``."""
        ...

    def size(self, i: Optional[int] = None) -> Union[int, torch.Size]:
        r"""Sampling grid size."""
        size = self.size_tensor()
        if i is None:
            return torch.Size(int(n) for n in size)
        return int(size[i])

    @property
    def shape(self) -> torch.Size:
        r"""Sampling grid size for dimensions ordered as ``(..., X)``."""
        return torch.Size((int(n) for n in self.size_tensor().flip(0)))

    def extent(self, i: Optional[int] = None) -> Tensor:
        r"""Extent of sampling grid in physical world units."""
        if i is None:
            return self.spacing() * self.size_tensor()
        return self._spacing[i] * self.size(i)

    def cube_extent(self, i: Optional[int] = None) -> Tensor:
        r"""Extent of sampling grid cube in physical world units."""
        if i is None:
            n = self.size_tensor()
            if self._align_corners:
                n = n.sub(1)
            return self.spacing().mul(n)
        n = self.size(i)
        if self._align_corners:
            n -= 1
        return self._spacing[i] * n

    @overload
    def center(self) -> Tensor:
        r"""Get grid center point in world space."""
        ...

    @overload
    def center(self, arg: Union[float, Array], *args: float) -> Grid:
        r"""Get new grid with specified center point in world space."""
        ...

    def center(self, arg: Union[float, Array, None] = None, *args: float) -> Union[Tensor, Grid]:
        r"""Get center point in world space or new grid with specified center point, respectively."""
        if arg is None:
            if args:
                raise ValueError(
                    f"{type(self).__name__}.center() 'args' cannot be given when first 'arg' is None"
                )
            return self._center
        return shallow_copy(self).center_(arg, *args)

    def center_(self, arg: Union[float, Array], *args: float) -> Grid:
        r"""Set grid center point in world space."""
        self._center = cat_scalars(arg, *args, num=self.ndim, dtype=self.dtype, device=self.device)
        return self

    @overload
    def origin(self) -> Tensor:
        r"""Get world coordinates of grid point with index zero."""
        ...

    @overload
    def origin(self, arg: Union[float, Array], *args: float) -> Grid:
        r"""Get new grid with specified world coordinates of grid point at index zero."""
        ...

    def origin(self, arg: Union[float, Array, None] = None, *args: float) -> Union[Tensor, Grid]:
        r"""Get grid origin in world space or new grid with specified origin, respectively."""
        if arg is None:
            if args:
                raise ValueError(
                    f"{type(self).__name__}.origin() 'args' cannot be given when first 'arg' is None"
                )
            size = self.size_tensor()
            offset = torch.where(size.gt(0), size.sub(1), size).div(2)
            offset = torch.matmul(self.affine(), offset)
            return self._center.sub(offset)
        return shallow_copy(self).origin_(arg, *args)

    def origin_(self, arg: Union[float, Array], *args: float) -> Grid:
        r"""Set world coordinates of grid point with zero index."""
        origin = cat_scalars(arg, *args, num=self.ndim, dtype=self.dtype, device=self.device)
        size = self.size_tensor()
        offset = torch.where(size.gt(0), size.sub(1), size).div(2)
        offset = torch.matmul(self.affine(), offset)
        self._center = origin.add(offset)
        return self

    @overload
    def spacing(self) -> Tensor:
        r"""Get spacing between grid points in world units."""
        ...

    @overload
    def spacing(self, arg: Union[float, Array], *args: float) -> Grid:
        r"""Get new grid with specified spacing between grid points in world units."""
        ...

    def spacing(self, arg: Union[float, Array, None] = None, *args: float) -> Union[Tensor, Grid]:
        r"""Get spacing between grid points in world units or new grid with specified spacing, respectively."""
        if arg is None:
            if args:
                raise ValueError(
                    f"{type(self).__name__}.spacing() 'args' cannot be given when first 'arg' is None"
                )
            return self._spacing
        return shallow_copy(self).spacing_(arg, *args)

    def spacing_(self, arg: Union[float, Array], *args: float) -> Grid:
        r"""Set spacing between grid points in physical world units."""
        spacing = cat_scalars(arg, *args, num=self.ndim, dtype=self.dtype, device=self.device)
        if spacing.le(0).any():
            raise ValueError("Grid spacing must be positive")
        self._spacing = spacing
        return self

    @overload
    def direction(self) -> Tensor:
        r"""Get grid axes direction cosines matrix."""
        ...

    @overload
    def direction(self, arg: Union[float, Array], *args: float) -> Grid:
        r"""Get new grid with specified axes direction cosines."""
        ...

    def direction(self, arg: Union[float, Array, None] = None, *args: float) -> Union[Tensor, Grid]:
        r"""Get grid axes direction cosines matrix or new grid with specified direction, respectively."""
        if arg is None:
            if args:
                raise ValueError(
                    f"{type(self).__name__}.direction() 'args' cannot be given when first 'arg' is None"
                )
            return self._direction
        return shallow_copy(self).direction_(arg, *args)

    def direction_(self, arg: Union[float, Array], *args: float) -> Grid:
        r"""Set grid axes direction cosines matrix of this grid."""
        D = self.ndim
        direction = torch.tensor((arg,) + args) if args else as_tensor(arg)
        direction = direction.to(dtype=self.dtype, device=self.device)
        if direction.ndim == 1:
            if direction.shape[0] != D * D:
                raise ValueError(
                    f"Grid direction must be array or square matrix with numel={D * D}"
                )
            direction = direction.reshape(D, D)
        else:
            if (
                direction.ndim != 2
                or direction.shape[0] != direction.shape[1]
                or direction.shape[0] != D
            ):
                raise ValueError(
                    f"Grid direction must be array or square matrix with numel={D * D}"
                )
        with torch.no_grad():
            if abs(direction.det().abs().item() - 1) > 1e-4:
                raise ValueError("Grid direction cosines matrix must be valid rotation matrix")
        self._direction = direction
        return self

    def affine(self) -> Tensor:
        r"""Affine transformation from ``Axes.GRID`` to ``Axes.WORLD``, excluding translation of origin."""
        return torch.mm(self.direction(), torch.diag(self.spacing()))

    def inverse_affine(self) -> Tensor:
        r"""Affine transformation from ``Axes.WORLD`` to ``Axes.GRID``, excluding translation of origin."""
        one = torch.tensor(1, dtype=self.dtype, device=self.device)
        return torch.mm(torch.diag(one / self.spacing()), self.direction().t())

    def transform(
        self,
        axes: Optional[Union[Axes, str]] = None,
        to_axes: Optional[Union[Axes, str]] = None,
        to_grid: Optional[Grid] = None,
        vectors: bool = False,
    ) -> Tensor:
        r"""Transformation from one grid domain to another.

        Args:
            axes: Axes with respect to which input coordinates are defined.
                If ``None`` and also ``to_axes`` and ``to_cube`` is ``None``,
                returns the transform which maps from cube to world space.
            to_axes: Axes of grid to which coordinates are mapped. Use ``axes`` if ``None``.
            to_grid: Other grid. Use ``self`` if ``None``.
            vectors: Whether transformation is used to rescale and reorient vectors.

        Returns:
            If ``vectors=False``, a homogeneous coordinate transformation of shape ``(D, D + 1)``.
            Otherwise, a square transformation matrix of shape ``(D, D)`` is returned.

        """
        if axes is None and to_axes is None and to_grid is None:
            cube_axes = Axes.CUBE_CORNERS if self._align_corners else Axes.CUBE
            return self.transform(cube_axes, Axes.WORLD, vectors=vectors)
        if axes is None:
            raise ValueError(
                "Grid.transform() 'axes' required when 'to_axes' or 'to_grid' specified"
            )
        matrix = None
        axes = Axes(axes)
        to_axes = axes if to_axes is None else Axes(to_axes)
        if to_grid is None or to_grid == self:
            if axes is to_axes:
                matrix = torch.eye(self.ndim, dtype=self.dtype, device=self.device)
                if not vectors:
                    offset = torch.zeros(self.ndim, dtype=self.dtype, device=self.device)
                    matrix = homogeneous_matrix(matrix, offset=offset)
            elif axes is Axes.GRID:
                if to_axes is Axes.CUBE:
                    size = self.size_tensor()
                    matrix = torch.diag(2 / size)
                    if not vectors:
                        one = torch.tensor(1, dtype=size.dtype, device=size.device)
                        matrix = homogeneous_matrix(matrix, offset=one / size - one)
                elif to_axes is Axes.CUBE_CORNERS:
                    size = self.size_tensor()
                    matrix = torch.diag(2 / (size - 1))
                    if not vectors:
                        offset = torch.tensor(-1, dtype=size.dtype, device=size.device)
                        matrix = homogeneous_matrix(matrix, offset=offset)
                elif to_axes is Axes.WORLD:
                    matrix = self.affine()
                    if not vectors:
                        matrix = homogeneous_matrix(matrix, offset=self.origin())
            elif axes is Axes.CUBE:
                if to_axes is Axes.CUBE_CORNERS:
                    size = self.size_tensor()
                    matrix = torch.diag(size / (size - 1))
                elif to_axes is Axes.GRID:
                    half_size = 0.5 * self.size_tensor()
                    matrix = torch.diag(half_size)
                    if not vectors:
                        matrix = homogeneous_matrix(matrix, offset=half_size - 0.5)
                elif to_axes is Axes.WORLD:
                    cube_to_grid = self.transform(axes, Axes.GRID, vectors=vectors)
                    grid_to_world = self.transform(Axes.GRID, Axes.WORLD, vectors=vectors)
                    if vectors:
                        matrix = torch.mm(grid_to_world, cube_to_grid)
                    else:
                        matrix = hmm(grid_to_world, cube_to_grid)
            elif axes is Axes.CUBE_CORNERS:
                if to_axes is Axes.CUBE:
                    size = self.size_tensor()
                    matrix = torch.diag((size - 1) / size)
                elif to_axes is Axes.GRID:
                    scales = 0.5 * (self.size_tensor() - 1)
                    matrix = torch.diag(scales)
                    if not vectors:
                        matrix = homogeneous_matrix(matrix, offset=scales)
                elif to_axes is Axes.WORLD:
                    interim = Axes.GRID
                    cube_to_grid = self.transform(axes, interim, vectors=vectors)
                    grid_to_world = self.transform(interim, to_axes, vectors=vectors)
                    if vectors:
                        matrix = torch.mm(grid_to_world, cube_to_grid)
                    else:
                        matrix = hmm(grid_to_world, cube_to_grid)
            elif axes is Axes.WORLD:
                if to_axes is Axes.CUBE or to_axes is Axes.CUBE_CORNERS:
                    interim = Axes.GRID
                    world_to_grid = self.transform(axes, interim, vectors=vectors)
                    grid_to_cube = self.transform(interim, to_axes, vectors=vectors)
                    if vectors:
                        matrix = torch.mm(grid_to_cube, world_to_grid)
                    else:
                        matrix = hmm(grid_to_cube, world_to_grid)
                elif to_axes is Axes.GRID:
                    matrix = self.inverse_affine()
                    if not vectors:
                        matrix = hmm(matrix, -self.origin())
        elif to_grid.ndim != self.ndim:
            raise ValueError(f"Grid.transform() 'to_grid' must have {self.ndim} spatial dimensions")
        else:
            target_to_world = self.transform(axes, Axes.WORLD, vectors=vectors)
            world_to_source = to_grid.transform(Axes.WORLD, to_axes, vectors=vectors)
            if vectors:
                matrix = torch.mm(world_to_source, target_to_world)
            else:
                matrix = hmm(world_to_source, target_to_world)
        if matrix is None:
            raise NotImplementedError(f"Grid.transform() for axes={axes} and to_axes={to_axes}")
        return matrix

    def inverse_transform(self, vectors: bool = False) -> Tensor:
        r"""Transform which maps from world to grid cube space."""
        cube_axes = Axes.CUBE_CORNERS if self._align_corners else Axes.CUBE
        return self.transform(Axes.WORLD, cube_axes, vectors=vectors)

    def apply_transform(
        self,
        input: Array,
        axes: Axes,
        to_axes: Optional[Axes] = None,
        to_grid: Optional[Grid] = None,
        vectors: bool = False,
        decimals: Optional[int] = -1,
    ) -> Tensor:
        r"""Map point coordinates or displacement vectors from one grid to another.

        Args:
            input: Points or vectors to transform as tensor of shape ``(..., D)``.
            axes: Axes with respect to which input coordinates are defined.
            to_axes: Axes of cube to which coordinates are mapped. Use ``axes`` if ``None``.
            to_cube: Other cube. Use ``self`` if ``None``.
            vectors: Whether transformation is used to rescale and reorient vectors.
            decimals: If positive or zero, number of digits right of the decimal point to round to.
                When mapping points to ``Axes.GRID``, ``Axes.CUBE``, or ``Axes.CUBE_CORNERS``,
                this function by default (``decimals=-1``) rounds the transformed coordinates.
                Explicitly set ``decimals=None`` to suppress this default rounding.

        Returns:
            If ``vectors=False``, a homogeneous coordinate transformation of shape ``(D, D + 1)``.
            Otherwise, a square transformation matrix of shape ``(D, D)`` is returned.

        """
        axes = Axes(axes)
        to_axes = axes if to_axes is None else Axes(to_axes)
        input = as_tensor(input)
        if not input.is_floating_point():
            input = input.type(self.dtype)
        if (to_grid is not None and to_grid != self) or axes is not to_axes:
            matrix = self.transform(axes, to_axes, to_grid=to_grid, vectors=vectors)
            matrix = matrix.unsqueeze(0).to(device=input.device)
            result = homogeneous_transform(matrix, input)
        else:
            result = input
        if decimals == -1:
            if to_axes is Axes.CUBE or to_axes is Axes.CUBE_CORNERS:
                decimals = 12
            elif to_axes is Axes.GRID:
                decimals = 6
        if decimals is not None and decimals >= 0:
            result = round_decimals(result, decimals=decimals)
        return result

    def transform_points(
        self,
        points: Array,
        axes: Axes,
        to_axes: Optional[Axes] = None,
        to_grid: Optional[Grid] = None,
        decimals: Optional[int] = -1,
    ) -> Tensor:
        r"""Map point coordinates from one grid domain to another.

        Args:
            points: Coordinates of points to transform as tensor of shape ``(..., D)``.
            axes: Coordinate axes with respect to which ``points`` are defined.
            to_axes: Coordinate axes to which ``points`` should be mapped to. If ``None``, use ``axes``.
            to_grid: Grid with respect to which the codomain is defined. If ``None``, the target
                and source sampling grids are assumed to be identical.
            decimals: If positive or zero, number of digits right of the decimal point to round to.
                When mapping points to codomain ``Axes.GRID``, ``Axes.CUBE``, or ``Axes.CUBE_CORNERS``,
                this function by default (``decimals=-1``) rounds the transformed coordinates.
                Explicitly set ``decimals=None`` to suppress this default rounding.

        Returns:
            Point coordinates in ``axes`` mapped to coordinates with respect to ``to_axes``.

        """
        return self.apply_transform(points, axes, to_axes, to_grid=to_grid, decimals=decimals)

    def transform_vectors(
        self,
        vectors: Array,
        axes: Axes,
        to_axes: Optional[Axes] = None,
        to_grid: Optional[Grid] = None,
    ) -> Tensor:
        r"""Rescale and reorient flow vectors.

        Args:
            vectors: Displacement vectors to transform, e.g., as tensor of shape ``(..., D)``.
            axes: Coordinate axes with respect to which ``vectors`` are defined.
            to_axes: Coordinate axes to which ``vectors`` should be mapped to. If ``None``, use ``axes``.
            to_grid: Grid with respect to which ``to_axes`` is defined. If ``None``, the target
                and source sampling grids are assumed to be identical.

        Returns:
            Rescaled and reoriented vectors. If ``to_grid == self`` and ``to_axes == axes``,
            a reference to the unmodified input ``vectors`` tensor is returned.

        """
        axes = Axes(axes)
        to_axes = axes if to_axes is None else Axes(to_axes)
        vectors = as_tensor(vectors)
        if not vectors.is_floating_point():
            vectors = vectors.type(self.dtype)
        if axes is Axes.WORLD and to_axes is Axes.WORLD:
            return vectors
        if to_grid is None or to_grid == self:
            if axes is not to_axes:
                affine = None  # affine transform required if reorientation needed
                scales = None  # otherwise, just scaling of displacements suffices
                if axes is Axes.WORLD:
                    affine = self.inverse_affine()
                elif axes is Axes.CUBE:
                    scales = self.size_tensor() / 2
                elif axes is Axes.CUBE_CORNERS:
                    scales = (self.size_tensor() - 1) / 2
                elif axes is not Axes.GRID:
                    raise NotImplementedError(
                        f"Grid.transform_vectors() for axes={axes} and to_axes={to_axes}"
                    )
                if to_axes is Axes.WORLD:
                    grid_to_world = self.affine()
                    if scales is None:
                        assert affine is None
                        affine = grid_to_world
                    else:
                        affine = torch.mm(grid_to_world, torch.diag(scales))
                elif to_axes is Axes.CUBE or to_axes is Axes.CUBE_CORNERS:
                    num = self.size_tensor()
                    if to_axes is Axes.CUBE_CORNERS:
                        num -= 1
                    grid_to_cube = 2 / num
                    if affine is None:
                        if scales is None:
                            scales = grid_to_cube
                        else:
                            scales *= grid_to_cube
                    else:
                        affine = torch.mm(torch.diag(grid_to_cube), affine)
                elif to_axes is not Axes.GRID:
                    raise NotImplementedError(
                        f"Grid.transform_vectors() for axes={axes} and to_axes={to_axes}"
                    )
                if affine is None:
                    assert scales is not None
                    scales = scales.to(vectors)
                    vectors = vectors * scales
                else:
                    affine = affine.to(vectors)
                    tensor = vectors.reshape(-1, vectors.shape[-1])
                    vectors = F.linear(tensor, affine).reshape(vectors.shape)
        else:
            matrix = self.transform(axes, to_axes, to_grid=to_grid, vectors=True)
            matrix = matrix.to(vectors)
            vectors = homogeneous_transform(matrix, vectors, vectors=True)
        return vectors

    def index_to_cube(
        self, indices: Array, decimals: int = -1, align_corners: Optional[bool] = None
    ) -> Tensor:
        r"""Map points from grid indices to grid-aligned cube with side length 2.

        Args:
            indices: Grid point indices to transform as tensor of shape ``(..., D)``.
            decimals: If positive or zero, number of digits right of the decimal point to round to.
            align_corners: Whether output cube coordinates should be with respect to
                ``Axes.CUBE_CORNERS`` (True) or ``Axes.CUBE`` (False), respectively.
                If ``None``, use default ``self.align_corners()``.

        Returns:
            Grid point indices transformed to points with respect to cube.

        """
        if align_corners is None:
            align_corners = self._align_corners
        to_axes = Axes.from_align_corners(align_corners)
        return self.transform_points(indices, axes=Axes.GRID, to_axes=to_axes, decimals=decimals)

    def cube_to_index(
        self, coords: Array, decimals: int = -1, align_corners: Optional[bool] = None
    ) -> Tensor:
        r"""Map points from grid-aligned cube to grid point indices.

        Args:
            coords: Normalized grid points to transform as tensor of shape ``(..., D)``.
            decimals: If positive or zero, number of digits right of the decimal point to round to.
            align_corners: Whether ``coords`` are with respect to ``Axes.CUBE_CORNERS`` (True)
                or ``Axes.CUBE`` (False), respectively. If ``None``, use default ``self.align_corners()``.

        Returns:
            Points in grid-aligned cube transformed to grid indices.

        """
        if align_corners is None:
            align_corners = self._align_corners
        axes = Axes.from_align_corners(align_corners)
        return self.transform_points(coords, axes=axes, to_axes=Axes.GRID, decimals=decimals)

    def index_to_world(self, indices: Array, decimals: int = -1) -> Tensor:
        r"""Map points from grid indices to world coordinates.

        Args:
            indices: Grid point indices to transform as tensor of shape ``(..., D)``.
            decimals: If positive or zero, number of digits right of the decimal point to round to.

        Returns:
            Grid point indices transformed to points in world space.

        """
        return self.transform_points(indices, axes=Axes.GRID, to_axes=Axes.WORLD, decimals=decimals)

    def world_to_index(self, points: Array, decimals: int = -1) -> Tensor:
        r"""Map points from world coordinates to grid point indices.

        Args:
            points: World coordinates of points to transform as tensor of shape ``(..., D)``.
            decimals: If positive or zero, number of digits right of the decimal point to round to.

        Returns:
            Points in world space transformed to grid indices.

        """
        return self.transform_points(points, axes=Axes.WORLD, to_axes=Axes.GRID, decimals=decimals)

    def cube_to_world(
        self, coords: Array, decimals: int = -1, align_corners: Optional[bool] = None
    ) -> Tensor:
        r"""Map point coordinates from grid-aligned cube with side length 2 to world space.

        Args:
            coords: Normalized grid points to transform as tensor of shape ``(..., D)``.
            decimals: If positive or zero, number of digits right of the decimal point to round to.
            align_corners: Whether ``coords`` are with respect to ``Axes.CUBE_CORNERS`` (True)
                or ``Axes.CUBE`` (False), respectively. If ``None``, use default ``self.align_corners()``.

        Returns:
            Normalized grid coordinates transformed to world space coordinates.

        """
        if align_corners is None:
            align_corners = self._align_corners
        axes = Axes.from_align_corners(align_corners)
        return self.transform_points(coords, axes=axes, to_axes=Axes.WORLD, decimals=decimals)

    def world_to_cube(
        self, points: Array, decimals: int = -1, align_corners: Optional[bool] = None
    ) -> Tensor:
        r"""Map point coordinates from world space to grid-aligned cube with side length 2.

        Args:
            points: World coordinates of points to transform as tensor of shape ``(..., D)``.
            decimals: If positive or zero, number of digits right of the decimal point to round to.
            align_corners: Whether output cube coordinates should be with respect to
                ``Axes.CUBE_CORNERS`` (True) or ``Axes.CUBE`` (False), respectively.
                If ``None``, use default ``self.align_corners()``.

        Returns:
            Points in world space transformed to normalized grid coordinates.

        """
        if align_corners is None:
            align_corners = self._align_corners
        to_axes = Axes.from_align_corners(align_corners)
        return self.transform_points(points, axes=Axes.WORLD, to_axes=to_axes, decimals=decimals)

    def coords(
        self,
        dim: Optional[int] = None,
        center: bool = False,
        normalize: bool = True,
        align_corners: Optional[bool] = None,
        channels_last: bool = True,
        flip: bool = False,
        dtype: Optional[DType] = None,
        device: Optional[Device] = None,
    ) -> Tensor:
        r"""Get tensor of grid point coordinates.

        Args:
            dim: Return coordinates for specified dimension, where ``dim=0`` refers to
                the first grid dimension, i.e., the ``x`` axis.
            center: Whether to center ``Axes.GRID`` coordinates when ``normalize=False``.
            normalize: Normalize coordinates to grid-aligned cube with side length 2.
                The normalized grid point coordinates are in the closed interval ``[-1, +1]``.
            align_corners: If ``normalize=True``, specifies whether the extrema ``(-1, 1)``
                should refer to the centers of the grid corner squares or their boundary.
                Note that in both cases the returned normalized coordinates are associated
                with the center points of the grid squares. When used as ``grid`` argument
                of ``torch.nn.functional.grid_sample()``, the same ``align_corners`` value
                should be used for both ``Grid.coords()`` and ``grid_sample()`` calls.
                If ``None``, use default ``self.align_corners()``.
            channels_last: Whether to place stacked coordinates at last (``True``) or first (``False``)
                output tensor dimension. Vector fields are represented with channels first after the
                batch dimension, whereas point sets such as the sampling points passed to ``grid_sample``
                are represented with point coordinates in the last tensor dimension.
            flip: Whether to return coordinates in the order (..., x) instead of (x, ...).
            dtype: Data type of coordinates. If ``None``, uses ``torch.int32`` as data type
                for returned tensor if ``normalize=False``, and ``self.dtype`` otherwise.
            device: Device on which to create PyTorch tensor. If ``None``, use ``self.device``.

        Returns:
            If ``dim`` is ``None``, returns a tensor of shape (...X, C) if ``channels_last=True`` (default)
            or ``(C, ..., X)`` if ``channels_last=False``, where C is the number of spatial grid dimensions.
            If ``normalize=Falze`` and ``center=False``, the tensor values are the multi-dimensional indices
            in the closed-open interval [0, n) for each grid dimension, where n is the number of points in the
            respective dimension. The first channel with index 0 is the ``x`` coordinate. If ``normalize=False``
            and ``center=True``, the indices are shifted such that index 0 corresponds to the grid center point.
            If ``normalize=True``, the centered coordinates are normalized to ``(-1, 1)``, where the extrema
            either correspond to the corner points of the grid (``align_corners=True``) or the grid boundary
            edges (``align_cornes=False``). If ``dim`` is not ``None``, a 1D tensor with the coordinates for
            this grid axis is returned.

        """
        if align_corners is None:
            align_corners = self._align_corners
        if dtype is None:
            if normalize or center:
                dtype = self.dtype
            else:
                dtype = torch.int32
        if device is None:
            device = self.device
        if dim is None:
            shape = self.shape  # order (...X), do NOT use self.size() here!
        else:
            shape = torch.Size((self.size()[dim],))  # dim may be negative
        if shape.numel() == 0:
            return torch.empty(shape, dtype=dtype, device=device)
        coords = []
        for n in shape:
            if n == 1:
                coord = torch.tensor([0], dtype=dtype, device=device)
            elif normalize:
                if align_corners:
                    spacing = 2 / (n - 1)
                    extrema = (-1, 1 + 0.1 * spacing)
                else:
                    spacing = 2 / n
                    extrema = (-1 + 0.5 * spacing, 1)
                coord = torch.arange(*extrema, spacing, dtype=dtype, device=device)
            elif center:
                radius = (n - 1) / 2
                coord = torch.linspace(-radius, radius, steps=n, dtype=dtype, device=device)
            else:
                coord = torch.arange(n, dtype=dtype, device=device)
            coords.append(coord)
        channels_dim = len(coords) if channels_last else 0
        if parse_version(torch.__version__) < parse_version("1.10"):
            coords = torch.stack(torch.meshgrid(*coords), dim=channels_dim)
        else:
            coords = torch.stack(torch.meshgrid(*coords, indexing="ij"), dim=channels_dim)
        if not flip:  # default order (x, ...) requires flipping stacked meshgrid coords
            coords = torch.flip(coords, dims=(channels_dim,))
        return coords

    def points(
        self,
        axes: Axes = Axes.WORLD,
        dtype: Optional[DType] = None,
        device: Optional[Device] = None,
    ) -> Tensor:
        r"""Tensor of grid point coordinates with respect to specified coordinate axes."""
        axes = Axes(axes)
        coords = self.coords(
            normalize=(axes is Axes.CUBE),
            align_corners=False,
            dtype=dtype,
            device=device,
        )
        if axes is not Axes.CUBE and axes is not Axes.GRID:
            coords = self.apply_transform(coords, Axes.GRID, to_axes=axes)
        return coords

    def _resize(self, size: Tensor, align_corners: Optional[bool] = None) -> Grid:
        r"""Get new grid of specified size.

        The resulting ``grid``` MUST preserve floating point values of given ``size`` argument such
        as in particular those passed by ``downsample()`` and ``upsample()``. Otherwise, a sequence
        of resampling steps which should produce the original grid may result in a different size.

        Args:
            size: Size of new grid.
            align_corners: Whether to preserve positions of corner points (``True``) or grid extent (``False``).

        """
        if align_corners is None:
            align_corners = self._align_corners
        size = size.to(dtype=self.dtype, device=self.device)
        if size.eq(self._size).all():
            return self
        grid = shallow_copy(self)
        grid._size = size
        size = self._round_size(size)
        if align_corners:
            spacing = (self.extent() - self.spacing()) / (size - 1)
            grid._spacing = torch.where(self._size.gt(0), spacing, self._spacing)
            assert torch.allclose(grid.origin(), self.origin())
        else:
            spacing = self.extent() / size
            grid._spacing = torch.where(self._size.gt(0), spacing, self._spacing)
            assert torch.allclose(grid.extent(), self.extent())
        return grid

    def resize(
        self,
        size: Union[int, Size, Array],
        *args: int,
        align_corners: Optional[bool] = None,
    ) -> Grid:
        r"""Create new grid of same extent with specified size.

        Specify new grid size for grid dimensions in the order (X...). Note that this is in reverse order
        of ``Tensor.size()``! To set the grid size given a data tensor shape, use ``Grid.reshape()``.

        Args:
            size: Size of new grid in the order ``(X, ...)``.
            align_corners: Whether to preserve positions of corner points (``True``) or grid extent (``False``).

        Returns:
            New grid with given ``size`` and adjusted ``Grid.spacing()``.

        """
        size = cat_scalars(size, *args, num=self.ndim, device=self.device)
        if not is_int_dtype(size.dtype):
            raise TypeError(f"Grid.resize() 'size' must be integer values, got dtype={size.dtype}")
        if size.lt(0).any():
            raise ValueError("Grid.resize() 'size' must be all non-negative numbers")
        return self._resize(size, align_corners=align_corners)

    def reshape(
        self,
        shape: Union[int, Shape, Array],
        *args: int,
        align_corners: Optional[bool] = None,
    ) -> Grid:
        r"""Create new grid of same extent with specified data tensor shape.

        The data tensor shape specifies the size of data dimensions in the order ``(..., X)``,
        whereas the ``Grid.size()`` is given in reverse order ``(X, ...)``. This function is a
        convenience function to change the grid size given the ``Tensor.shape`` of a data tensor.

        Args:
            shape: Size of new grid in the order ``(..., X)``.
            align_corners: Whether to preserve positions of corner points (``True``) or grid extent (``False``).

        Returns:
            New grid with given ``shape`` and adjusted ``Grid.spacing()``.

        """
        shape = cat_scalars(shape, *args, num=self.ndim, device=self.device)
        if not is_int_dtype(shape.dtype):
            raise TypeError(
                f"Grid.reshape() 'shape' must be integer values, got dtype={shape.dtype}"
            )
        if shape.lt(0).any():
            raise ValueError("Grid.reshape() 'shape' must be all non-negative numbers")
        return self._resize(shape.flip(0), align_corners=align_corners)

    def resample(self, spacing: Union[float, Array, str], *args: float, min_size: int = 1) -> Grid:
        r"""Create new grid with specified spacing.

        Args:
            spacing: Desired spacing between grid points. Uses minimum or maximum grid spacing
                for isotropic resampling when argument is string "min" or "max", respectively.
            min_size: Minimum grid size.

        Returns:
            New grid with specified spacing. The extent of the grid may be greater
            than before, if the original extent is not divisible by the desired spacing.

        """
        if spacing == "min":
            assert not args
            spacing = self._spacing.min()
        elif spacing == "max":
            assert not args
            spacing = self._spacing.max()
        elif isinstance(spacing, str):
            raise ValueError(
                f"{type(self).__name__}.resample() 'spacing' str must be 'min' or 'max'"
            )
        spacing = cat_scalars(spacing, *args, num=self.ndim, dtype=self.dtype, device=self.device)
        if torch.allclose(spacing, self._spacing):
            return self
        if spacing.le(0).any():
            raise ValueError("Grid.resample() 'spacing' must be all positive numbers")
        size = self.extent().div(spacing)
        size = torch.where(self._size.gt(0), size.clamp(min=min_size), size)
        grid = shallow_copy(self)
        grid._size = size
        grid._spacing = spacing
        return grid

    def pool(
        self,
        kernel_size: ScalarOrTuple[int],
        stride: Optional[ScalarOrTuple[int]] = None,
        padding: ScalarOrTuple[int] = 0,
        dilation: ScalarOrTuple[int] = 1,
        ceil_mode: bool = False,
    ) -> Grid:
        r"""Output grid after applying pooling operation.

        Args:
            kernel_size: Size of the pooling region.
            stride: Stride of the pooling operation.
            padding: Implicit zero paddings on both sides of the input.
            dilation: Spacing between pooling kernel elements.
            ceil_mode: When True, will use `ceil` instead of `floor` to compute the output size.

        Returns:
            New grid corresponding to output data tensor after pooling operation.

        """
        if stride is not None:
            raise NotImplementedError("Grid.pool() 'stride' currently not supported")
        if padding != 0:
            raise NotImplementedError("Grid.pool() 'padding' currently not supported")
        if dilation != 1:
            raise NotImplementedError("Grid.pool() 'dilation' currently not supported")
        ks = cat_scalars(kernel_size, num=self.ndim, dtype=self.dtype, device=self.device)
        size = self.size_tensor() / ks
        size = size.ceil() if ceil_mode else size.floor()
        size = size.type(torch.int)
        grid = Grid(
            size=size,
            origin=self.index_to_world(ks.sub(1).div(2)),
            spacing=self.spacing().mul(ks),
            direction=self.direction(),
            align_corners=self.align_corners(),
            device=self.device,
        )
        return grid

    def avg_pool(
        self,
        kernel_size: ScalarOrTuple[int],
        stride: Optional[ScalarOrTuple[int]] = None,
        padding: ScalarOrTuple[int] = 0,
        ceil_mode: bool = False,
    ) -> Grid:
        r"""Output grid after applying average pooling."""
        return self.pool(kernel_size, stride=stride, padding=padding, ceil_mode=ceil_mode)

    def downsample(
        self,
        levels: int = 1,
        dims: Optional[Sequence[SpatialDimArg]] = None,
        min_size: int = 1,
        align_corners: Optional[bool] = None,
    ) -> Grid:
        r"""Create new grid with size halved the specified number of times.

        Args:
            levels: Number of times the grid size is halved (>0) or doubled (<0).
            dims: Spatial dimensions along which to downsample. If not specified, consider all spatial dimensions.
            min_size: Minimum grid size. If downsampling the grid along a spatial dimension would reduce its
                size below the given minimum value, the grid is not further downsampled along this dimension.
            align_corners: Whether to preserve positions of corner points (``True``) or grid extent (``False``).

        Returns:
            New grid.

        """
        if not isinstance(levels, int):
            raise TypeError("Grid.downsample() 'levels' must be of type int")
        if not dims:
            dims = tuple(dim for dim in range(self.ndim))
        dims = tuple(SpatialDim.from_arg(dim) for dim in dims)
        size = self._size.clone()
        scale = 2**levels
        for dim in dims:
            size[dim] /= scale
        size = torch.where(size.ge(min_size), size, self._size)
        return self._resize(size, align_corners=align_corners)

    def upsample(
        self,
        levels: int = 1,
        dims: Optional[Sequence[SpatialDimArg]] = None,
        align_corners: Optional[bool] = None,
    ) -> Grid:
        r"""Create new grid of same extent with size doubled the specified number of times.

        Args:
            levels: Number of times the grid size is doubled (>0) or halved (<0).
            dims: Spatial dimensions along which to upsample. If not specified, consider all spatial dimensions.
            min_size: Minimum grid size. If downsampling the grid along a spatial dimension would reduce its
                size below the given minimum value, the grid is not further downsampled along this dimension.
            align_corners: Whether to preserve positions of corner points (``True``) or grid extent (``False``).

        Returns:
            New grid.

        """
        if not isinstance(levels, int):
            raise TypeError("Grid.upsample() 'levels' must be of type int")
        if not dims:
            dims = tuple(dim for dim in range(self.ndim))
        dims = tuple(SpatialDim.from_arg(dim) for dim in dims)
        size = self._size.clone()
        scale = 2**levels
        for dim in dims:
            size[dim] *= scale
        return self._resize(size, align_corners=align_corners)

    def pyramid(
        self,
        levels: int,
        dims: Optional[Sequence[SpatialDimArg]] = None,
        min_size: int = 0,
    ) -> Dict[int, Grid]:
        r"""Compute image size at each image resolution pyramid level.

        This function computes suitable image sizes for each level of a multi-resolution
        image pyramid depending on original image size and spacing, a minimum grid size for
        every level, and whether grid corners (``align_corners=True``) or grid borders
        (``align_corners=False``) should be aligned.

        Args:
            levels: Number of resolution levels.
            dims: Spatial dimensions along which to downsample. If not specified, consider all spatial dimensions.
            min_size: Minimum grid size at each level. If the grid size after downsampling
                would be smaller than the specified minimum size, the grid size is not reduced
                for this spatial dimension. The number of resolution levels is not affected.

        Returns:
            Dictionary mapping resolution level to sampling grid. The sampling grid at the
            finest resolution level has index 0. The cube extent, i.e., the physical length
            between grid points corresponding to cube interval ``(-1, 1)`` will be the same
            for all resolution levels.

        """
        if not isinstance(levels, int):
            raise TypeError("Grid.pyramid() 'levels' must be int")
        if not isinstance(min_size, int):
            raise TypeError("Grid.pyramid() 'min_size' must be int")
        if not dims:
            dims = tuple(dim for dim in range(self.ndim))
        dims = tuple(SpatialDim.from_arg(dim) for dim in dims)
        m = sum([2**i for i in range(levels)]) if self._align_corners else 0
        sizes = {level: list(self.size()) for level in range(levels + 1)}
        for dim in dims:
            sizes[levels][dim] = int(0.5 + (sizes[levels][dim] + m) / (2**levels))
            for level in range(levels - 1, -1, -1):
                sizes[level][dim] = 2 * sizes[level + 1][dim] - 1
            for level in range(1, levels + 1):
                sizes[level][dim] = (sizes[level - 1][dim] + 1) // 2
                if sizes[level][dim] < min_size:
                    sizes[level][dim] = sizes[level - 1][dim]
        return {level: self.resize(size) for level, size in sizes.items()}

    def crop(
        self,
        *args: int,
        margin: Optional[Union[int, Array]] = None,
        num: Optional[Union[int, Array]] = None,
    ) -> Grid:
        r"""Create new grid with a margin along each axis removed.

        Args:
            args: Crop ``margin`` specified as int arguments.
            margin: Number of spatial grid points to remove (positive) or add (negative) at each border.
                Use instead of ``num`` in order to symmetrically crop the input ``data`` tensor, e.g.,
                ``(nx, ny, nz)`` is equivalent to ``num=(nx, nx, ny, ny, nz, nz)``.
            num: Number of spatial grid points to remove (positive) or add (negative) at each border,
                where margin of the last dimension of the ``data`` tensor must be given first, e.g.,
                ``(nx, nx, ny, ny)``. If a scalar is given, the input is cropped equally at all borders.
                Otherwise, the given sequence must have an even length.

        Returns:
            New grid with modified ``Grid.size()``, but unchanged ``Grid.spacing``.
            Hence, the ``Grid.extent()`` of the new grid will be different from ``self.extent()``.

        """
        if sum([1 if args else 0, 0 if num is None else 1, 0 if margin is None else 1]) != 1:
            raise AssertionError("Grid.pad() 'args', 'margin', and 'num' are mutually exclusive")
        if len(args) == 1 and not isinstance(args[0], int):
            margin = args[0]
        elif args:
            margin = args
        if isinstance(margin, int):
            num = margin
        elif margin is not None:
            num = tuple(n for n_n in ((n, n) for n in margin) for n in n_n)
        assert num is not None
        if isinstance(num, int):
            num = (num,) * (2 * self.ndim)
        else:
            num = tuple(int(n) for n in num)
        if len(num) % 2 != 0:
            raise ValueError("Grid.crop() 'num' must be int or have even length")
        if all(n == 0 for n in num):
            return self
        num = num + (0,) * (2 * self.ndim - len(num))
        num_ = torch.tensor(num, dtype=self.dtype, device=self.device)
        size = torch.clamp(self._size - num_[::2] - num_[1::2], min=1)
        size = torch.where(self._size.gt(0), size, self._size)
        origin = self.index_to_world(num_[::2])
        return Grid(
            size=size,
            origin=origin,
            spacing=self.spacing(),
            direction=self.direction(),
            align_corners=self.align_corners(),
            device=self.device,
        )

    def pad(
        self,
        *args: int,
        margin: Optional[Union[int, Array]] = None,
        num: Optional[Union[int, Array]] = None,
    ) -> Grid:
        r"""Create new grid with an additional margin along each axis.

        Args:
            args: Pad ``margin`` specified as int arguments.
            margin: Number of spatial grid points to add (positive) or remove (negative) at each border.
                Use instead of ``num`` in order to symmetrically pad the input ``data`` tensor, e.g.,
                ``(nx, ny, nz)`` is equivalent to ``num=(nx, nx, ny, ny, nz, nz)``.
            num: Number of spatial grid points to remove (positive) or add (negative) at each border,
                where margin of the last dimension of the ``data`` tensor must be given first, e.g.,
                ``(nx, nx, ny, ny)``. If a scalar is given, the input is cropped equally at all borders.
                Otherwise, the given sequence must have an even length.

        Returns:
            New grid with modified ``Grid.size()``, but unchanged ``Grid.spacing``.
            Hence, the ``Grid.extent()`` of the new grid will be different from ``self.extent()``.

        """
        if sum([1 if args else 0, 0 if num is None else 1, 0 if margin is None else 1]) != 1:
            raise AssertionError("Grid.pad() 'args', 'margin', and 'num' are mutually exclusive")
        if len(args) == 1 and not isinstance(args[0], int):
            margin = args[0]
        elif args:
            margin = args
        if isinstance(margin, int):
            num = margin
        elif margin is not None:
            num = tuple(n for n_n in ((n, n) for n in margin) for n in n_n)
        assert num is not None
        if isinstance(num, int):
            num = (num,) * (2 * self.ndim)
        else:
            num = tuple(int(n) for n in num)
        if len(num) % 2 != 0:
            raise ValueError("Grid.pad() 'num' must be int or have even length")
        if all(n == 0 for n in num):
            return self
        num = num + (0,) * (2 * self.ndim - len(num))
        num_ = torch.tensor(num, dtype=self.dtype, device=self.device)
        size = torch.clamp(self._size + num_[::2] + num_[1::2], min=1)
        size = torch.where(self._size.gt(0), size, self._size)
        origin = self.index_to_world(-num_[::2])
        return Grid(
            size=size,
            origin=origin,
            spacing=self.spacing(),
            direction=self.direction(),
            align_corners=self.align_corners(),
            device=self.device,
        )

    def center_crop(self, size: Union[int, Array], *args: int) -> Grid:
        r"""Crop grid to specified maximum size."""
        size = cat_scalars(size, *args, num=self.ndim, device=self.device)
        if not is_int_dtype(size.dtype):
            raise TypeError(
                f"Grid.center_crop() expected scalar or array of integer values, got dtype={size.dtype}"
            )
        size = [min(m, n) for m, n in zip(self.size(), size.tolist())]
        origin = [(m - n) // 2 for m, n in zip(self.size(), size)]
        return Grid(
            size=size,
            origin=self.index_to_world(origin),
            spacing=self.spacing(),
            direction=self.direction(),
            align_corners=self.align_corners(),
            device=self.device,
        )

    def center_pad(self, size: Union[int, Array], *args: int) -> Grid:
        r"""Pad grid to specified minimum size."""
        size = cat_scalars(size, *args, num=self.ndim, device=self.device)
        if not is_int_dtype(size.dtype):
            raise TypeError(
                f"Grid.center_crop() expected scalar or array of integer values, got dtype={size.dtype}"
            )
        size = [max(m, n) for m, n in zip(self.size(), size.tolist())]
        origin = [-((n - m) // 2) for m, n in zip(self.size(), size)]
        return Grid(
            size=size,
            origin=self.index_to_world(origin),
            spacing=self.spacing(),
            direction=self.direction(),
            align_corners=self.align_corners(),
            device=self.device,
        )

    def narrow(self, dim: int, start: int, length: int) -> Grid:
        r"""Narrow grid along specified dimension."""
        if dim < 0 or dim > self.ndim:
            raise IndexError("Grid.narrow() 'dim' is out of bounds")
        size = tuple(length if d == dim else n for d, n in enumerate(self.size()))
        origin = tuple(start if d == dim else 0 for d in range(self.ndim))
        return Grid(
            size=size,
            origin=self.index_to_world(origin),
            spacing=self.spacing(),
            direction=self.direction(),
            align_corners=self.align_corners(),
            device=self.device,
        )

    def region_of_interest(self, start: Union[int, Array], size: Union[int, Array]) -> Grid:
        r"""Get region of interest grid."""
        start = cat_scalars(start, num=self.ndim, device=self.device)
        if not is_int_dtype(start.dtype):
            raise TypeError(
                f"Grid.region_of_interest() 'start' must be scalar or array of integer values, got dtype={start.dtype}"
            )
        size = cat_scalars(size, num=self.ndim, device=self.device)
        if not is_int_dtype(size.dtype):
            raise TypeError(
                f"Grid.region_of_interest() 'size' must be scalar or array of integer values, got dtype={size.dtype}"
            )
        grid_size = self.size()
        num = [[start[i], grid_size[i] - (start[i] + size[i])] for i in range(self.ndim)]
        num = [n for nn in num for n in nn]
        return self.crop(num=num)

    def same_domain_as(self, other: Grid) -> bool:
        """Check if this and another grid cover the same cube domain."""
        if other is self:
            return True
        return self.domain() == other.domain()

    def __eq__(self, other: Any) -> bool:
        r"""Compare this grid to another."""
        if other is self:
            return True
        if not isinstance(other, self.__class__):
            return False
        for name in self.__slots__:
            if name == "_align_corners":
                continue
            value = getattr(self, name)
            other_value = getattr(other, name)
            if type(value) != type(other_value):
                return False
            if isinstance(value, Tensor):
                assert isinstance(other_value, Tensor)
                if value.shape != other_value.shape:
                    return False
                other_value = other_value.to(device=value.device)
                if not torch.allclose(value, other_value, rtol=1e-5, atol=1e-8):
                    return False
            elif value != other_value:
                return False
        return True

    def __repr__(self) -> str:
        r"""String representation."""
        size = ", ".join([f"{v:>6.2f}" for v in self._size])
        center = ", ".join([f"{v:.5f}" for v in self._center])
        origin = ", ".join([f"{v:.5f}" for v in self.origin()])
        spacing = ", ".join([f"{v:.5f}" for v in self._spacing])
        direction = ", ".join([f"{v:.5f}" for v in self._direction.flatten()])
        return (
            f"{type(self).__name__}("
            + f"size=({size})"
            + f", origin=({origin})"
            + f", center=({center})"
            + f", spacing=({spacing})"
            + f", direction=({direction})"
            + f", device={repr(str(self.device))}"
            + f", align_corners={repr(self._align_corners)}"
            + ")"
        )


def grid_points_transform(grid: Grid, axes: Axes, to_grid: Grid, to_axes: Optional[Axes] = None):
    r"""Get linear transformation of points from one grid domain to another.

    Args:
        grid: Sampling grid with respect to which input points are defined.
        axes: Grid axes with respect to which input points are defined.
        to_grid: Sampling grid with respect to which output points are defined.
        to_axes: Grid axes with respect to which output points are defined.

    Returns:
        Homogeneous coordinate transformation matrix as tensor of shape ``(D, D + 1)``.

    """
    return grid.transform(axes=axes, to_axes=to_axes, to_grid=to_grid, vectors=False)


def grid_vectors_transform(grid: Grid, axes: Axes, to_grid: Grid, to_axes: Optional[Axes] = None):
    r"""Get affine transformation which maps vectors with respect to one grid domain to another.

    Args:
        grid: Sampling grid with respect to which input vectors are defined.
        axes: Grid axes with respect to which input vectors are defined.
        to_grid: Sampling grid with respect to which output vectors are defined.
        to_axes: Grid axes with respect to which output vectors are defined.

    Returns:
        Affine transformation matrix as tensor of shape ``(D, D)``.

    """
    return grid.transform(axes=axes, to_axes=to_axes, to_grid=to_grid, vectors=True)


def grid_transform_points(
    points: Tensor,
    grid: Grid,
    axes: Axes,
    to_grid: Grid,
    to_axes: Optional[Axes] = None,
    decimals: Optional[int] = -1,
):
    r"""Map point coordinates from one grid domain to another.

    Args:
        points: Coordinates of points to transform as tensor of shape ``(..., D)``.
        grid: Grid with respect to which input ``points`` are defined.
        axes: Coordinate axes with respect to which ``points`` are defined.
        to_grid: Grid with respect to which the codomain is defined. If ``None``, the target
            and source sampling grids are assumed to be identical.
        to_axes: Coordinate axes to which ``points`` should be mapped to. If ``None``, use ``axes``.
        decimals: If positive or zero, number of digits right of the decimal point to round to.
            When mapping points to codomain ``Axes.GRID``, ``Axes.CUBE``, or ``Axes.CUBE_CORNERS``,
            this function by default (``decimals=-1``) rounds the transformed coordinates.
            Explicitly set ``decimals=None`` to suppress this default rounding.

    Returns:
        Point coordinates in ``axes`` mapped to coordinates with respect to ``to_axes``.

    """
    return grid.transform_points(
        points, axes=axes, to_axes=to_axes, to_grid=to_grid, decimals=decimals
    )


def grid_transform_vectors(
    vectors: Tensor,
    grid: Grid,
    axes: Axes,
    to_grid: Grid,
    to_axes: Optional[Axes] = None,
):
    r"""Rescale and reorient flow vectors.

    Args:
        vectors: Displacement vectors to transform, e.g., as tensor of shape ``(..., D)``.
        grid: Grid with respect to which input ``vectors`` are defined.
        axes: Coordinate axes with respect to which input ``vectors`` are defined.
        to_grid: Grid with respect to which ``to_axes`` is defined. If ``None``, the target
            and source sampling grids are assumed to be identical.
        to_axes: Coordinate axes to which ``vectors`` should be mapped to. If ``None``, use ``axes``.

    Returns:
        Rescaled and reoriented vectors. If ``to_grid == grid`` and ``to_axes == axes``,
        a reference to the unmodified input ``vectors`` tensor is returned.

    """
    return grid.transform_vectors(vectors, axes=axes, to_axes=to_axes, to_grid=to_grid)



r"""Auxiliary functions for working with vtkImageData."""

from typing import Optional

import numpy as np

from vtk import (
    vtkImageData,
    vtkImageStencilData,
    vtkImageStencilToImage,
    vtkMatrixToLinearTransform,
    vtkPolyData,
    vtkPolyDataToImageStencil,
    vtkTransformPolyDataFilter,
)

from deepali.core.grid import Grid

from .numpy import numpy_to_vtk_matrix4x4


def surface_mesh_grid(*mesh: vtkPolyData, resolution: Optional[float] = None) -> Grid:
    r"""Compute image grid for given surface mesh discretization with specified resolution."""
    # Common bounds of mesh points
    #
    # ATTENTION: vtkBooleanOperationPolyDataFilter is too unreliable to use!
    # (e.g., cf. http://vtk.1045678.n5.nabble.com/vtkBooleanOperationPolyDataFilter-crashes-while-trying-to-intersect-2-vtkTubeFilter-outputs-td5746999.html)
    bounds = np.zeros((6,), dtype=float)
    for pointset in mesh:
        _bounds = np.asarray(pointset.GetBounds())
        bounds[0::2] = np.minimum(bounds[0::2], _bounds[0::2])
        bounds[1::2] = np.maximum(bounds[1::2], _bounds[1::2])
    # Calculate default resolution
    if resolution is None or resolution <= 0 or np.isnan(resolution):
        resolution = np.sqrt(np.sum(np.square(bounds[1::2] - bounds[0::2]))) / 256
    # Calculate grid properties for bounding box
    return Grid(
        size=np.ceil((bounds[1::2] - bounds[0::2]) / resolution).astype(int),
        origin=bounds[0::2] + 0.5 * resolution,
        spacing=np.asarray([resolution] * 3),
    )


def surface_image_stencil(mesh: vtkPolyData, grid: Grid) -> vtkImageStencilData:
    r"""Convert vtkPolyData surface mesh to image stencil."""
    max_index = [n - 1 for n in grid.size().tolist()]

    rot = np.eye(4, dtype=np.float)
    rot[:3, :3] = np.array(grid.direction).reshape(3, 3)
    rot = numpy_to_vtk_matrix4x4(rot)

    transform = vtkMatrixToLinearTransform()
    transform.SetInput(rot)

    transformer = vtkTransformPolyDataFilter()
    transformer.SetInputData(mesh)
    transformer.SetTransform(transform)

    converter = vtkPolyDataToImageStencil()
    converter.SetInputConnection(transformer.GetOutputPort())
    converter.SetOutputOrigin(grid.origin().tolist())
    converter.SetOutputSpacing(grid.spacing().tolist())
    converter.SetOutputWholeExtent([0, max_index[0], 0, max_index[1], 0, max_index[2]])
    converter.Update()

    stencil = vtkImageStencilData()
    stencil.DeepCopy(converter.GetOutput())
    return stencil


def binary_image_stencil_mask(stencil: vtkImageStencilData) -> vtkImageData:
    r"""Set values inside image stencil to specified value."""
    converter = vtkImageStencilToImage()
    converter.SetInsideValue(1)
    converter.SetOutsideValue(0)
    converter.SetInputData(stencil)
    converter.SetOutputScalarTypeToUnsignedChar()
    converter.Update()
    mask = vtkImageData()
    mask.DeepCopy(converter.GetOutput())
    return mask



r"""Composite spatial transformations are composed of other spatial transforms."""

from __future__ import annotations

from copy import copy as shallow_copy
from collections import OrderedDict
from typing import Iterable, Optional, Tuple, TypeVar, Union, overload

import torch
from torch import Tensor
from torch.nn import ModuleDict

from deepali.core.grid import Axes, Grid, grid_transform_points
from deepali.core.linalg import as_homogeneous_matrix, homogeneous_matmul
from deepali.core.tensor import move_dim

from .base import SpatialTransform


TCompositeTransform = TypeVar("TCompositeTransform", bound="CompositeTransform")


class CompositeTransform(SpatialTransform):
    r"""Base class of composite spatial coordinate transformations.

    Base class of modules that apply one or more spatial transformations to map a tensor of
    spatial points to another tensor of spatial points of the same shape as the input tensor.

    """

    @overload
    def __init__(self, grid: Grid) -> None:
        r"""Initialize empty composite transformation."""
        ...

    @overload
    def __init__(self, grid: Grid, *args: Optional[SpatialTransform]) -> None:
        r"""Initialize composite transformation."""
        ...

    @overload
    def __init__(self, grid: Grid, transforms: Union[OrderedDict, ModuleDict]) -> None:
        r"""Initialize composite transformation given named transforms in ordered dictionary."""
        ...

    @overload
    def __init__(self, *args: Optional[SpatialTransform]) -> None:
        r"""Initialize composite transformation."""
        ...

    @overload
    def __init__(self, transforms: Union[ModuleDict, OrderedDict]) -> None:
        r"""Initialize composite transformation given named transforms in ordered dictionary."""
        ...

    def __init__(
        self, *args: Optional[Union[Grid, ModuleDict, OrderedDict, SpatialTransform]]
    ) -> None:
        r"""Initialize composite transformation."""
        args_ = [arg for arg in args if arg is not None]
        grid = None
        if isinstance(args_[0], Grid):
            grid = args_[0]
            args_ = args_[1:]
        if args_:
            if isinstance(args_[0], (dict, ModuleDict)):
                if len(args_) > 1:
                    raise ValueError(
                        f"{type(self).__name__}() multiple arguments not allowed when dict is given"
                    )
                transforms = args_[0]
            else:
                transforms = OrderedDict([(str(i), t) for i, t in enumerate(args_)])
        else:
            transforms = OrderedDict()
        if grid is None:
            if transforms:
                transform = next(iter(transforms.values()))
                grid = transform.grid()
            else:
                raise ValueError(
                    f"{type(self).__name__}() requires a Grid or at least one SpatialTransform"
                )
        for name, transform in transforms.items():
            if not isinstance(transform, SpatialTransform):
                raise TypeError(
                    f"{type(self).__name__}() module '{name}' must be of type SpatialTransform"
                )
            if not transform.grid().same_domain_as(grid):
                raise ValueError(
                    f"{type(self).__name__}() transform '{name}' has different 'grid' center, direction, or cube extent"
                )
        super().__init__(grid)
        self._transforms = ModuleDict(transforms)

    def bool(self) -> bool:
        r"""Whether this module has at least one transformation."""
        return len(self._transforms) > 0

    def __len__(self) -> int:
        r"""Number of spatial transformations."""
        return len(self._transforms)

    @property
    def linear(self) -> bool:
        r"""Whether composite transformation is linear."""
        return all(transform.linear for transform in self.transforms())

    def __contains__(self, name: Union[int, str]) -> bool:
        r"""Whether composite contains named transformation."""
        if isinstance(name, int):
            name = str(name)
        return name in self._transforms.keys()

    def __getitem__(self, name: Union[int, str]) -> SpatialTransform:
        r"""Get named transformation."""
        if isinstance(name, int):
            name = str(name)
        return self._transforms[name]

    def get(
        self, name: Union[int, str], default: Optional[SpatialTransform] = None
    ) -> Optional[SpatialTransform]:
        r"""Get named transformation."""
        if isinstance(name, int):
            name = str(name)
        for key, transform in self._transforms.items():
            if key == name:
                assert isinstance(transform, SpatialTransform)
                return transform
        return default

    def transforms(self) -> Iterable[SpatialTransform]:
        r"""Iterate transformations in order of composition."""
        return self._transforms.values()

    def named_transforms(self) -> Iterable[Tuple[str, SpatialTransform]]:
        r"""Iterate transformations in order of composition."""
        return self._transforms.items()

    def condition(
        self: TCompositeTransform, *args, **kwargs
    ) -> Union[TCompositeTransform, Optional[Tensor]]:
        r"""Get or set data tensor on which transformations are conditioned."""
        if args or kwargs:
            return shallow_copy(self).condition_(*args, **kwargs)
        return self._args, self._kwargs

    def condition_(self: TCompositeTransform, *args, **kwargs) -> TCompositeTransform:
        r"""Set data tensor on which transformations are conditioned."""
        assert args or kwargs
        super().condition_(*args, **kwargs)
        for transform in self.transforms():
            transform.condition_(*args, **kwargs)
        return self

    def disp(self, grid: Optional[Grid] = None) -> Tensor:
        r"""Get displacement vector field representation of this transformation.

        Args:
            grid: Grid on which to sample vector fields. Use ``self.grid()`` if ``None``.

        Returns:
            Displacement vector fields as tensor of shape ``(N, D, ..., X)``.

        """
        if grid is None:
            grid = self.grid()
        axes = Axes.from_grid(grid)
        x = grid.coords(device=self.device).unsqueeze(0)
        if grid.same_domain_as(self.grid()):
            y = self.forward(x)
        else:
            y = grid_transform_points(x, grid, axes, self.grid(), self.axes())
            y = self.forward(y)
            y = grid_transform_points(y, self.grid(), self.axes(), grid, axes)
        u = y - x
        u = move_dim(u, -1, 1)
        return u

    def update(self: TCompositeTransform) -> TCompositeTransform:
        r"""Update buffered data such as predicted parameters, velocities, and/or displacements."""
        super().update()
        for transform in self.transforms():
            transform.update()
        return self

    def clear_buffers(self: TCompositeTransform) -> TCompositeTransform:
        r"""Clear any buffers that are registered by ``self.update()``."""
        super().clear_buffers()
        for transform in self.transforms():
            transform.clear_buffers()
        return self


class MultiLevelTransform(CompositeTransform):
    r"""Sum of spatial transformations applied to any set of points.

    A :class:`.MultiLevelTransform` adds the sum of the displacement vectors across all
    spatial transforms at the input points that are being mapped to new locations, i.e.,

    .. math::

        \vec{y} = \vec{x} + \sum_{i=0}^{n-1} \vec{u}_i(\vec{x})

    """

    def forward(self, points: Tensor, grid: bool = False) -> Tensor:
        r"""Transform set of points by sum of spatial transformations.

        Args:
            points: Tensor of shape ``(N, M, D)`` or ``(N, ..., Y, X, D)``.
            grid: Whether ``points`` are the positions of undeformed grid points.

        Returns:
            Tensor of same shape as ``points`` with transformed point coordinates.

        """
        x = points
        if len(self) == 0:
            return x
        if self.linear:
            # The base class implementation uses composite self.tensor(), which is
            # more efficient in case of the composition of linear transformations.
            y = super().forward(points, grid)
        else:
            u = torch.zeros_like(x)
            for i, transform in enumerate(self.transforms()):
                y = transform.forward(x, grid=grid and i == 0)
                u += y - x
            y = x + u
        return y

    def tensor(self) -> Tensor:
        r"""Get tensor representation of this transformation.

        The tensor representation of a transformation is with respect to the unit cube axes defined
        by its sampling grid as specified by ``self.axes()``.

        Returns:
            In case of a composition of linear transformations, returns a batch of homogeneous transformation
            matrices as tensor of shape ``(N, D, 1)`` (translation),  ``(N, D, D)`` (affine) or ``(N, D, D + 1)``,
            i.e., a 3-dimensional tensor. If this composite transformation contains a non-rigid transformation,
            a displacement vector field is returned as tensor of shape ``(N, D, ..., X)``.

        """
        if self.linear:
            transforms = list(self.transforms())
            if not transforms:
                identity = torch.eye(self.ndim, self.ndim + 1, device=self.device)
                return identity.unsqueeze(0)
            transform = transforms[0]
            mat = as_homogeneous_matrix(transform.tensor())
            for transform in transforms[1:]:
                mat += as_homogeneous_matrix(transform.tensor())
            return mat
        return self.disp()


class SequentialTransform(CompositeTransform):
    r"""Composition of spatial transformations applied to any set of points.

    A :class:`.SequentialTransform` is the functional composition of spatial transforms, i.e.,

    .. math::

        \vec{y} = \vec{u}_{n-1} \circ \cdots \circ \vec{u}_0 \circ \vec{x}

    """

    def forward(self, points: Tensor, grid: bool = False) -> Tensor:
        r"""Transform points by sequence of spatial transformations.

        Args:
            points: Tensor of shape ``(N, M, D)`` or ``(N, ..., Y, X, D)``.
            grid: Whether ``points`` are the positions of undeformed grid points.

        Returns:
            Tensor of same shape as ``points`` with transformed point coordinates.

        """
        # The base class implementation uses self.tensor(), which is more efficient
        # in case of the composition of only linear transformations.
        if self.linear:
            return super().forward(points, grid)
        y = points
        for i, transform in enumerate(self.transforms()):
            y = transform.forward(y, grid=grid and i == 0)
        return y

    def tensor(self) -> Tensor:
        r"""Get tensor representation of this transformation.

        The tensor representation of a transformation is with respect to the unit cube axes defined
        by its sampling grid as specified by ``self.axes()``.

        Returns:
            In case of a composition of linear transformations, returns a batch of homogeneous transformation
            matrices as tensor of shape ``(N, D, 1)`` (translation),  ``(N, D, D)`` (affine) or ``(N, D, D + 1)``,
            i.e., a 3-dimensional tensor. If this composite transformation contains a non-rigid transformation,
            a displacement vector field is returned as tensor of shape ``(N, D, ..., X)``.

        """
        if self.linear:
            transforms = list(self.transforms())
            if not transforms:
                identity = torch.eye(self.ndim, self.ndim + 1, device=self.device)
                return identity.unsqueeze(0)
            transform = transforms[0]
            mat = transform.tensor()
            for transform in transforms[1:]:
                mat = homogeneous_matmul(transform.tensor(), mat)
            return mat
        return self.disp()

    def inverse(
        self: TCompositeTransform, link: bool = False, update_buffers: bool = False
    ) -> TCompositeTransform:
        r"""Get inverse of this transformation.

        Args:
            link: Whether to inverse transformation keeps a reference to this transformation.
                If ``True``, the ``update()`` function of the inverse function will not recompute
                shared parameters, e.g., parameters obtained by a callable neural network, but
                directly access the parameters from this transformation. Note that when ``False``,
                the inverse transformation will still share parameters, modules, and buffers with
                this transformation, but these shared tensors may be replaced by a call of ``update()``
                (which is implicitly called as pre-forward hook when ``__call__()`` is invoked).
            update_buffers: Whether buffers of inverse transformation should be update after creating
                the shallow copy. If ``False``, the ``update()`` function of the returned inverse
                transformation has to be called before it is used.

        Returns:
            Shallow copy of this transformation which computes and applied the inverse transformation.
            The inverse transformation will share the parameters with this transformation. Not all
            transformations may implement this functionality.

        Raises:
            NotImplementedError: When a transformation does not support sharing parameters with its inverse.

        """
        copy = shallow_copy(self)
        transforms = ModuleDict()
        for name, transform in reversed(self.named_transforms()):
            assert isinstance(transform, SpatialTransform)
            transforms[name] = transform.inverse(link=link, update_buffers=update_buffers)
        copy._transforms = transforms
        return copy
